package mc.construction;

import mc.model.Joint;
import mc.model.BasicJoint;
import mc.model.AnimationModel;
import mc.model.SkeletonHandle;
import mc.model.ModelListener;
import mc.model.ModelEvent;

import com.sun.java.util.collections.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.tree.*;

/**
 * The construction view
 */
class ConstructionView extends JTree implements ModelListener
{
    /** The animation model */
    private AnimationModel theModel;

    /** The tree model */
    private DefaultTreeModel treeModel;
    
    /** The default tree node */
    private DefaultMutableTreeNode root;
    
    /** The id counter */
    private static int idCounter;

    /** A map from skeletons to nodes */
    private final Map nodeMap;
    
    /**
     * Initialize the view
     *
     */
    public ConstructionView()
    {
	super();
	nodeMap=new HashMap();
	theModel=null;
	root=new DefaultMutableTreeNode("Skeletons");
	root.setUserObject(new JointData(0,null,null));
	initTreeModel(root);
    }
    
    /**
     * Set the animation model for the view.
     *
     * @param model The animation model
     */
    public void setAnimationModel(AnimationModel model)
    {
	if (theModel!=null) {
	    theModel.removeModelListener(this);
	}
	
	theModel=model;
	nodeMap.clear();
	
	if (theModel!=null) {
	    theModel.addModelListener(this);
	    buildTreeModel();
	    repaint();
	}
    }

    /**
     * Build the tree model.
     *
     */
    public void buildTreeModel()
    {
	// Create the root node
	root=new DefaultMutableTreeNode();
	root.setUserObject(new JointData(-1,null,null));
	
	// Iterate over the skeletons in the original model
	for(Iterator i=theModel.getHandles().iterator();i.hasNext();) {
	    SkeletonHandle handle=(SkeletonHandle)i.next();
	    // Add it to the root node
	    buildLODSkeletonHierarchy(handle);
	}
	
	// Init the tree model
	initTreeModel(root);
    }

    private void buildLODSkeletonHierarchy(SkeletonHandle handle)
    {
	// Iterate over the LOD Pyramid
	for(Iterator i=handle.getLODPyramid().getLevels();i.hasNext();) {
	    SkeletonHandle lodHandle=(SkeletonHandle)i.next();
	    // Build the tree
	    root.add(buildSkeletonHierarchy(lodHandle,lodHandle.getRootJoint()));
	}
    }
    
    private void initTreeModel(DefaultMutableTreeNode root)
    {
	// Create the model
	treeModel=new DefaultTreeModel(root);
	setModel(treeModel);
	// Create the selection model
	DefaultTreeSelectionModel selectionModel=new DefaultTreeSelectionModel();
	selectionModel.setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
	setSelectionModel(selectionModel);
	// Set the renderer
	JointNodeRenderer renderer=new JointNodeRenderer();
	renderer.setClosedIcon(renderer.getDefaultLeafIcon());
	renderer.setOpenIcon(renderer.getDefaultLeafIcon());
	setCellRenderer(renderer);
	// Make sure all rows are expanded
	for(int i=0;i<getRowCount();i++) {
	    expandRow(i);
	}
    }

    /**
     * Add another skeleton to the view.
     *
     * @param treeNode The skeleton.
     */
    public void addSkeleton(MutableTreeNode treeNode,SkeletonHandle skeleton)
    {
	root.add(treeNode);
	treeModel.reload();
	nodeMap.put(skeleton,treeNode);
	repaint();
    }
    
    /**
     * Build a skeleton hierarchy.
     *
     * @return The tree node for the root of the hierarchy
     */
    private DefaultMutableTreeNode buildSkeletonHierarchy(SkeletonHandle handle,Joint joint)
    {
	DefaultMutableTreeNode treeNode=new DefaultMutableTreeNode();
	treeNode.setUserObject(new JointData(idCounter++,joint,handle));
	// Iterate over the children
	for(Iterator i=joint.getChildren().iterator();i.hasNext();) {
	    Joint childJoint=(Joint)i.next();
	    // Recurse
	    treeNode.add(buildSkeletonHierarchy(handle,childJoint));
	}
	return treeNode;
    }

    public void frameTicked(int frame,int increment)
    {}
    
    public void modelModified(ModelEvent e)
    {
	// Check the type of modification
	if (e.getType()==ModelEvent.SKELETON_ADDED) {
	    // Get the new skeleton
	    SkeletonHandle skeleton=e.getHandle();
	    // Build the new tree nodes
	    DefaultMutableTreeNode newNode=buildSkeletonHierarchy(skeleton,skeleton.getRootJoint());
	    addSkeleton(newNode,skeleton);
	} 
	else if (e.getType()==ModelEvent.SKELETON_CHANGED) {
	    // Get the modified skeleton
	    SkeletonHandle skeleton=e.getHandle();
	    // Find the root node of the handle
	    DefaultMutableTreeNode treeNode=(DefaultMutableTreeNode)nodeMap.get(skeleton);
	    // Get its index
	    int index=root.getIndex(treeNode);
	    // Build a new hierarchy
	    DefaultMutableTreeNode newNode=buildSkeletonHierarchy(skeleton,skeleton.getRootJoint());
	    // Replace the old one
	    root.insert(newNode,index);
	    root.remove(treeNode);
	    nodeMap.put(skeleton,newNode);
	    // Reload and repaint
	    treeModel.reload();
	    repaint();
	}
    }

    /**
     * Get the tree node that is associated with a skeleton.
     *
     * @param handle The handle for the skeleton.
     * @return the root of the skeleton in the tree.
     */
    public DefaultMutableTreeNode getTreeNodeFor(SkeletonHandle handle)
    {
	return (DefaultMutableTreeNode)nodeMap.get(handle);
    }
}

/**
 * A renderer for joints
 */
class JointNodeRenderer extends DefaultTreeCellRenderer
{
    public Component getTreeCellRendererComponent(JTree tree, Object value, boolean sel, boolean expanded, boolean leaf, int row, boolean hasFocus) 
    {
	super.getTreeCellRendererComponent(tree,value,sel,expanded,leaf,row,hasFocus);
	// Get the node
	DefaultMutableTreeNode node=(DefaultMutableTreeNode)value;
	// Get the pair
	JointData data=(JointData)node.getUserObject();
	// Set the text
	if (data.theJoint==null) {
	    setText("Skeletons");
	} else {
	    setText(data.theJoint.getName());
	}
	return this;
    }

}

