package mc.construction;

import mc.model.AnimationModel;
import mc.io.bvh.BVHModelReader;
import mc.model.Joint;
import mc.model.SkeletonHandle;
import mc.model.ModelTimer;
import mc.model.BasicAnimationModel;
import mc.model.DriverFactory;
import mc.model.euler.EulerAnglesDriverFactory;
import mc.model.ModelException;
import mc.model.BasicJoint;
import mc.model.TranslationDriver;
import mc.model.ModelEvent;
import mc.model.MappedTranslationDriver;
import mc.model.euler.MappedEulerRotationDriver;
import mc.model.euler.EulerRotationDriver;
import mc.model.DataManager;
import mc.model.DataRepresentation;
import mc.model.ModelEvent;
import mc.view.TranslationVectorPanel;
import mc.model.TranslationVector3d;
import mc.util.FormatBigDecimal;
import mc.model.LODPyramid;
import mc.model.DefaultLODChooser;

import com.sun.java.util.collections.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import javax.vecmath.*;

/**
 * The skeleton editor
 */
public class SkeletonEditor extends JPanel implements ActionListener, TreeSelectionListener, ItemListener
{
    /** The view for the offsets */
    private TranslationVectorPanel xlatView;
    
   /** The collapse button */
    private JButton collapseButton;
    
    /** The clone button */
    private JButton cloneButton;
    
    /** The visible check box */
    private JCheckBox visibleCheckBox;
    
    /** The set of visible skeletons */
    private Set visible;
    
    /** The view */
    private ConstructionView theView;
    
    /** the model */
    private BasicAnimationModel theModel;

    /** The handle to root handle map */
    private Map rootHandles;
    
    /** The node that was selected */
    private JointData selected;

    /** The root handle */
    private SkeletonHandle theRootHandle;
    
    /**
     * The clone command 
     */
    private final static String CLONE="CLONE";
    
    /**
     * The collapse command
     */
    private final static String COLLAPSE="COLLAPSE";

    /**
     * The elevation command.
     */
    private final static String ELEVATION="ELEVATION";
    
    /**
     * The elevation field.
     */
    private JTextField elevationField;
    
    /**
     * The map that holds the data about a skeleton.
     */
    private Map skeletonData;
    
    /**
     * Initialize.
     *
     */
    public SkeletonEditor()
    {
	this.skeletonData=new HashMap();
	this.visible=new HashSet();
	this.rootHandles=new HashMap();
	
	// Initialize the visual components
	initializeComponents();
    }

    /**
     * Return the animation model
     *
     */
    public BasicAnimationModel getAnimationModel()
    {
	return theModel;
    }
    
    /**
     * Initialize the visual components
     *
     */
    private void initializeComponents()
    {
	// Set the layout
	GridBagLayout layout=new GridBagLayout();
	setLayout(layout);
	GridBagConstraints c=new GridBagConstraints();
	
	// Build the tree view
	theView=new ConstructionView();
	theView.setVisibleRowCount(10);
	theView.addTreeSelectionListener(this);
	JScrollPane pane=new JScrollPane(theView);
	pane.setBorder(BorderFactory.createTitledBorder("Skeletons"));
	
	c.gridwidth=2;
	c.gridheight=4;
	c.weightx=c.weighty=1;
	c.insets=new Insets(5,5,5,5);
	c.fill=GridBagConstraints.BOTH;
	layout.setConstraints(pane,c);
	add(pane);
	
	// Build the buttons
	JPanel buttonPanel=new JPanel();
	GridBagLayout buttonLayout=new GridBagLayout();
	buttonPanel.setLayout(buttonLayout);
	buttonPanel.setBorder(BorderFactory.createTitledBorder("Controls"));
	
	c.gridwidth=GridBagConstraints.REMAINDER;
	c.gridheight=1;
	c.anchor=GridBagConstraints.NORTH;
	c.fill=GridBagConstraints.HORIZONTAL;
	c.weightx=0.5;
	c.weighty=0.5;
	c.anchor=GridBagConstraints.CENTER;
	layout.setConstraints(buttonPanel,c);
	add(buttonPanel);
	
	// The clone button
	cloneButton=new JButton("Clone");
	cloneButton.addActionListener(this);
	cloneButton.setActionCommand(CLONE);
	cloneButton.setEnabled(false);
	
	c.gridwidth=1;
	c.gridheight=1;
	c.gridx=c.gridy=GridBagConstraints.RELATIVE;
	c.fill=GridBagConstraints.HORIZONTAL;
	c.weighty=1;
	c.weightx=1;
	c.anchor=GridBagConstraints.CENTER;
	buttonLayout.setConstraints(cloneButton,c);
	buttonPanel.add(cloneButton);
	
	// Collapse button
	collapseButton=new JButton("Collapse");
	collapseButton.addActionListener(this);
	collapseButton.setActionCommand(COLLAPSE);
	collapseButton.setEnabled(false);
	
	c.gridwidth=GridBagConstraints.REMAINDER;
	c.gridheight=1;
	c.gridx=c.gridy=GridBagConstraints.RELATIVE;
	c.fill=GridBagConstraints.HORIZONTAL;
	c.weighty=1;
	c.weightx=1;
	c.anchor=GridBagConstraints.CENTER;
	buttonLayout.setConstraints(collapseButton,c);
	buttonPanel.add(collapseButton);

	// Build the elevation field
	elevationField=new JTextField(5);
	elevationField.addActionListener(this);
	elevationField.setActionCommand(ELEVATION);
	elevationField.setHorizontalAlignment(JTextField.LEFT);
	JPanel elevationPanel=new JPanel(new BorderLayout());
	elevationPanel.add(BorderLayout.WEST,new JLabel("Distance:"));
	elevationPanel.add(BorderLayout.CENTER,elevationField);
	elevationPanel.setBorder(BorderFactory.createTitledBorder("Elevation"));
	c.gridwidth=GridBagConstraints.REMAINDER;
	c.gridheight=1;
	c.gridx=c.gridy=GridBagConstraints.RELATIVE;
	c.fill=GridBagConstraints.NONE;
	c.weightx=0.5;
	c.weighty=0.5;
	c.anchor=GridBagConstraints.CENTER;
	layout.setConstraints(elevationPanel,c);
	add(elevationPanel);
	
	// Build the visible field
	visibleCheckBox=new JCheckBox("Show in View");
	visibleCheckBox.addItemListener(this);
	c.gridwidth=GridBagConstraints.REMAINDER;
	c.gridheight=1;
	c.gridx=c.gridy=GridBagConstraints.RELATIVE;
	c.fill=GridBagConstraints.NONE;
	c.weightx=0.5;
	c.weighty=0.5;
	c.anchor=GridBagConstraints.CENTER;
	layout.setConstraints(visibleCheckBox,c);
	add(visibleCheckBox);
	
	// Build the translation view
	JPanel xlatPanel=new JPanel(new BorderLayout());
	xlatView=new TranslationVectorPanel();
	xlatPanel.add(BorderLayout.CENTER,xlatView);
	
	c.gridwidth=GridBagConstraints.REMAINDER;
	c.gridheight=1;
	c.weightx=0.5;
	c.weighty=0.5;
	c.fill=GridBagConstraints.BOTH;
	c.anchor=GridBagConstraints.CENTER;
	layout.setConstraints(xlatPanel,c);
	add(xlatPanel);
    }

    /**
     * Collapse a given node, and connect
     * its children to its parent. Also takes care
     * of adding the offsets.
     *
     */
    private void collapseNode()
    {
	// Get the selected tree path
	TreePath selectedPath=theView.getSelectionPath();
	// Get the last element ...
	int numOfComponents=selectedPath.getPathCount();
	DefaultMutableTreeNode node=(DefaultMutableTreeNode)selectedPath.getLastPathComponent();
	JointData data=(JointData)node.getUserObject();
	// Get the data manager 
	DataManager dataManager=data.theHandle.getDataManager();
	// Get the joint and the offset
	BasicJoint joint=(BasicJoint)data.theJoint;
	Matrix4d matrix=data.theHandle.getDataManager().getOffset(data.theJoint);
	// Get the parent node
	DefaultMutableTreeNode parent=(DefaultMutableTreeNode)selectedPath.getPathComponent(numOfComponents-2);
	JointData parentData=(JointData)parent.getUserObject();
	// The parent joint
	BasicJoint parentJoint=(BasicJoint)parentData.theJoint;
	// The index of the joint in the parent joint
	int nodeIndex=parentJoint.getIndexOf(data.theJoint);
	// Reparent the children of the joint
	while(joint.getChildCount()!=0){
	    // Get the child
	    BasicJoint child=(BasicJoint)joint.getChildAt(0);
	    // Reparent the child
	    parentJoint.insertChild(child,nodeIndex);
	    nodeIndex++;
	    // Calculate a new translation matrix
	    Matrix4d childMatrix=dataManager.getOffset(child);
	    Matrix4d newMatrix=new Matrix4d();
	    newMatrix.mul(matrix,childMatrix);
	    // Send the matrix to the translation driver
	    Vector3d newOffset=new Vector3d();
	    newMatrix.get(newOffset);
	    dataManager.setOffset(child,newOffset.x,newOffset.y,newOffset.z);
	}
	// Remove the original node
	parentJoint.remove(joint);
	// Rebuild the view
	theModel.fireNotification(new ModelEvent(theModel,ModelEvent.SKELETON_CHANGED,data.theHandle));
    }
    
    /**
     * Set the animation model.
     *
     * @param model The animation model.
     * @param elevations The elevation map.
     */
    public void setAnimationModel(BasicAnimationModel model,Map elevations) 
    {
	skeletonData.clear();
	this.theModel=model;
	theView.setAnimationModel(theModel);
	// Iterate over the skeletons and assign a default elevation of 1
	for(Iterator i=model.getHandles().iterator();i.hasNext();) {
	    SkeletonHandle handle=(SkeletonHandle)i.next();
	    recordElevation(handle,elevations);
	    // Iterate over the LODs
	    for(Iterator li=handle.getLODPyramid().getLevels();li.hasNext();) {
		SkeletonHandle lodHandle=(SkeletonHandle)li.next();
		rootHandles.put(lodHandle,handle);
		recordElevation(lodHandle,elevations);
	    }
	}
    }

    /**
     * Record the elevation for a given skeleton. If no
     * elevation is specified, a default elevation of one is assigned.
     *
     * @param handle The handle.
     * @param elevations The elevations map.
     */
    private void recordElevation(SkeletonHandle handle,Map elevations)
    {
	// Check if the elevation is set
	Double elevation=(Double)elevations.get(handle);
	if (elevation==null) {
	    skeletonData.put(handle,new Double(1));
	} else {
	    skeletonData.put(handle,elevation);
	}
    }
	
    /**
     * A button was pressed.
     *
     */
    public void actionPerformed(ActionEvent e)
    {
	if (e.getActionCommand()==COLLAPSE) {
	    collapseNode();
	    expandAll();
	    resetSelection();
	}
	else if (e.getActionCommand()==CLONE) {
	    cloneSkeleton();
	}
	else if (e.getActionCommand()==ELEVATION) {
	    updateElevation();
	    elevationField.transferFocus();
	}
    }

    private void updateElevation()
    {
	// Get the selected node
	TreePath path=theView.getSelectionPath();
	// Check for boundary condition
	if (path.getPathCount()<2) {
	    elevationField.setText("");
	    return;
	}
	DefaultMutableTreeNode node=(DefaultMutableTreeNode)path.getPathComponent(1);
	JointData data=(JointData)node.getUserObject();
	// Check for root node
	if (data.theJoint==null) {
	    elevationField.setText("");
	    return;
	}
	
	// Try to parse the text from the field
	Double newElevation=null;
	try {
	    newElevation=new Double(elevationField.getText());
	}
	catch(NumberFormatException e) {
	    updateField(data.theHandle);
	    JOptionPane.showMessageDialog(getTopLevelAncestor(),"Error","Invalid elevation",JOptionPane.ERROR_MESSAGE);
	    return;
	}
	
	// Update the elevation - first get the hips joint
	skeletonData.put(data.theHandle,newElevation);
	updateField(data.theHandle);
    }

    /**
     * Update the elevation field with the
     * elevation of a skeleton.
     *
     * @param handle The skeleton handle.
     */
    private void updateField(SkeletonHandle handle)
    {
	Double d=(Double)skeletonData.get(handle);
	if (d==null) {
	    elevationField.setText("");
	} else {
	    elevationField.setText(FormatBigDecimal.format(d.doubleValue(),2).toString());
	}
    }

    /**
     * Get the elevation value for a skeleton.
     *
     * @param handle The handle.
     * @return The elevation or 1 if it was not set.
     */
    public double getElevationFor(SkeletonHandle handle)
    {
	Double d=(Double)skeletonData.get(handle);
	if (d==null) {
	    return 1;
	} else {
	    return d.doubleValue();
	}
    }
    
    /**
     * Make sure all rows are expanded.
     *
     */
    public void expandAll()
    {
	for(int i=0;i<theView.getRowCount();i++) {
	    theView.expandRow(i);
	}
    }
    
    /**
     * A selection was made.
     *
     */
    public void valueChanged(TreeSelectionEvent e)
    {
	// Get the path
	TreePath nodePath=e.getPath();
	// Get the node
	DefaultMutableTreeNode treeNode=(DefaultMutableTreeNode)nodePath.getLastPathComponent();
	// Get the joint data
	selected=(JointData)treeNode.getUserObject();
	// Check if the root is selected
	if (selected.theJoint==null) {
	    collapseButton.setEnabled(false);
	    cloneButton.setEnabled(false);
	    elevationField.setText("");
	    visibleCheckBox.setEnabled(false);
	    return;
	}
	// Update the channel view
	Matrix4d matrix=selected.theHandle.getDataManager().getOffset(selected.theJoint);
	Vector3d xlat=new Vector3d();
	matrix.get(xlat);
	DataRepresentation xlatRep=new TranslationVector3d(xlat.x,xlat.y,xlat.z);
	xlatView.refresh(xlatRep);
	
	updateField(selected.theHandle);

	// Update the visible check box
	visibleCheckBox.removeItemListener(this);
	visibleCheckBox.getModel().setSelected(visible.contains(selected.theHandle));
	visibleCheckBox.addItemListener(this);
	
	// Check if a skeleton root has been selected
	if (selected.theJoint.getParent()==null) {
	    cloneButton.setEnabled(true);
	    collapseButton.setEnabled(false);
	} else {
	    collapseButton.setEnabled(true);
	    cloneButton.setEnabled(false);
	}
    }

    /**
     * Clone a skeleton.
     *
     * @param root The root.
     */
    private void cloneSkeleton()
    {
	Map cloneMap=new HashMap();
	// Get the Joint at the end
	BasicJoint clone=clone(selected.theJoint,cloneMap);
	// Add the skeleton
	SkeletonHandle handle=theModel.createSkeleton(clone);
	SkeletonHandle theRootHandle=(SkeletonHandle)rootHandles.get(selected.theHandle);
	rootHandles.put(handle,theRootHandle);
	theRootHandle.getLODPyramid().addLevel(handle);
	// Notify the view
	theModel.fireNotification(new ModelEvent(this,ModelEvent.SKELETON_ADDED,handle));
	// Clone the offsets
	for(Iterator i=cloneMap.keySet().iterator();i.hasNext();) {
	    Joint cloneJoint=(Joint)i.next();
	    // Get the original joint
	    Joint origJoint=(Joint)cloneMap.get(cloneJoint);
	    // Get the offset from the original joint
	    Matrix4d offset=selected.theHandle.getDataManager().getOffset(origJoint);
	    // Set the offset
	    handle.getDataManager().setOffset(cloneJoint,new Matrix4d(offset));
	}
	// Clone the base drivers
	DataManager dataManager=handle.getDataManager();
	// Now select the clone
	expandAll();
	selectNode(handle);
    }
    
    /**
     * Select a particular node in the tree.
     *
     * @param handle the handle to select.
     */
    private void selectNode(SkeletonHandle handle)
    {
	DefaultMutableTreeNode node=theView.getTreeNodeFor(handle);
	DefaultTreeModel treeModel=(DefaultTreeModel)theView.getModel();
	TreePath path=new TreePath(treeModel.getPathToRoot(node));
	theView.setSelectionPath(path);
	theView.scrollPathToVisible(path);
    }
    
    /**
     * Clone a whole subtree.
     *
     * @param node The root of the tree.
     * @return The cloned subtree.
     */
    private BasicJoint clone(Joint joint,Map cloneMap)
    {
	// Create a new joint
	BasicJoint clone=new BasicJoint(joint);
	// Update the mapping
	cloneMap.put(clone,joint);
	// Now clone the children of the node
	for(Iterator i=joint.getChildren().iterator();i.hasNext();) {
	    Joint child=(Joint)i.next();
	    // Clone the child
	    BasicJoint cloneChild=clone(child,cloneMap);
	    // Add it to the cloned tree
	    clone.addChild(cloneChild);
	}
	return clone;
    }
    
    /**
     * Reset the selection of the user.
     *
     */
    private void resetSelection()
    {
	// Deselect the node in the tree
	theView.getSelectionModel().clearSelection();
	// Disable the buttons
	collapseButton.setEnabled(false);
	cloneButton.setEnabled(false);
    }

    public void itemStateChanged(ItemEvent e) 
    {
	if (e.getStateChange()==ItemEvent.SELECTED) {
	    visible.add(selected.theHandle);
	} else {
	    visible.remove(selected.theHandle);
	}
    }
    
}

class MyWindowAdapter extends WindowAdapter
{
    public void windowClosing(WindowEvent e)
    {
	System.exit(0);
    }
}


