package mc.construction;

import mc.io.bvh.LODBVHParser;
import mc.io.bvh.ParseException;
import mc.model.ModelException;
import mc.model.euler.EulerAnglesDriverFactory;
import mc.model.ModelTimer;
import mc.model.DriverFactory;
import mc.model.BasicAnimationModel;
import mc.model.LODAnimationModel;
import mc.model.SkeletonHandle;
import mc.model.Joint;
import mc.model.DataManager;
import mc.model.euler.EulerRotationDriver;
import mc.view.RenderableView;
import mc.view.GLModelView;

import javax.vecmath.Matrix4d;
import javax.vecmath.Vector3d;
import javax.vecmath.Tuple3d;
import com.sun.java.util.collections.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;

/**
 * The skeleton editor frame
 */
public class SkeletonEditorFrame extends JFrame implements ActionListener
{
    private final static String EXPORT="Export";
    
    private final static String LOAD="Load";
    
    private final static String EXIT="Exit";
    
    /** The editor component */
    private final SkeletonEditor editor;

    /** The file chooser */
    private final JFileChooser chooser;
    
    /** The view that will supply the camera */
    private final RenderableView theView;
    
    /**
     * The skeleton editor
     *
     */
    public SkeletonEditorFrame(RenderableView theView)
    {
	super();
	this.editor=new SkeletonEditor();
	this.chooser=new JFileChooser();
	this.theView=theView;
	
	// Initialize the componetns
	initComponents();
	// Time for the show
	setSize(new Dimension(300,400));
	pack();
	setVisible(true);
    }

    private void initComponents()
    {
	setContentPane(new JPanel(new BorderLayout()));
	getContentPane().add(BorderLayout.CENTER,editor);

	// Get the menu bar
	JMenuBar menuBar=new JMenuBar();
	// Add the file menu
	JMenu fileMenu=new JMenu("File");
	menuBar.add(fileMenu);
	getContentPane().add(BorderLayout.NORTH,menuBar);
	
	// Add the load item
	JMenuItem loadItem=new JMenuItem("Load");
	loadItem.addActionListener(this);
	loadItem.setActionCommand(LOAD);
	fileMenu.add(loadItem);
	// Add the export item
	JMenuItem exportItem=new JMenuItem("Export");
	exportItem.addActionListener(this);
	exportItem.setActionCommand(EXPORT);
	fileMenu.add(exportItem);
	// Add the separator
	fileMenu.add(new JSeparator());
	// Add the exit
	JMenuItem exitItem=new JMenuItem("Exit");
	exitItem.addActionListener(this);
	exitItem.setActionCommand(EXIT);
	fileMenu.add(exitItem);
    }
    
    public void actionPerformed(ActionEvent e)
    {
	if (e.getActionCommand().equals(LOAD)) {
	    loadFile();
	}
	else if (e.getActionCommand().equals(EXPORT)) {
	    exportFile();
	}
	else if (e.getActionCommand().equals(EXIT)) {
	    System.exit(0);
	} else {
	    throw new Error("Unknown command "+e.getActionCommand());
	}
    }

    /**
     * Present a load dialog to the user and load the appropriate file.
     *
     */
    private void loadFile()
    {
	// Present the file selector
	chooser.showOpenDialog(this);
	// Get the file
	File file=chooser.getSelectedFile();
	if (file==null) {
	    return ;
	}
	loadFile(file);
    }
    
    public void loadFile(File file)
    {
	// Read in the animation model
	ModelTimer timer=new ModelTimer(300);
	DriverFactory factory=new EulerAnglesDriverFactory();
	LODAnimationModel model=new LODAnimationModel(timer,theView);
	model.setDriverFactory(factory);
	
	Map elevations=new HashMap();
	try {
	    LODBVHParser parser=new LODBVHParser(new FileInputStream(file));
	    parser.init();
	    parser.fill(model,elevations);
	    parser.Hierarchy();
	}
	catch(IOException e) {
	    JOptionPane.showMessageDialog(this,"Could not open file "+file.getPath(),"Error",JOptionPane.ERROR_MESSAGE);
	    return;
	}
	catch(ParseException e) {
	    JOptionPane.showMessageDialog(this,"Error loading file: "+e.getMessage(),"Error",JOptionPane.ERROR_MESSAGE);
	    return;
	}
	
	// Set the model
	editor.setAnimationModel(model,elevations);
    }
    
    /**
     * Export the modified skeletons
     *
     */
    private void exportFile()
    {
	// Present the file selector
	chooser.showDialog(this,"Export");
	// Get the file
	File file=chooser.getSelectedFile();
	// Get a writer on the file
	try {
	    Writer writer=new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file)));
	    export(editor.getAnimationModel(),writer);
	    writer.flush();
	    writer.close();
	}
	catch(IOException e) {
	    JOptionPane.showMessageDialog(this,"Cannot open file: "+e.getMessage(),"Error",JOptionPane.ERROR_MESSAGE);
	}
    }

    /**
     * Export the motion part of an animation model.
     *
     * @param model The basic animation model.
     * @param writer The writer
     * @exception IOException if an error occurs
     */
    private void exportMotion(BasicAnimationModel model,Writer writer) throws IOException
    {
	writer.write("Frames: "+model.getNumberOfFrames()+"\n");
	writer.write("Frame Time: "+(1/(double)30)+"\n");
	// Replay the frames
	for(int frame=0;frame<model.getNumberOfFrames();frame++) {
	    // Iterate over the skeletons
	    boolean first=true;
	    for(Iterator i=model.getHandles().iterator();i.hasNext();) {
		// Get the skeleton
		SkeletonHandle skeleton=(SkeletonHandle)i.next();
		writeMotion(skeleton,frame,writer,true);
		// Write the levels of detail
		for(Iterator li=skeleton.getLODPyramid().getLevels();li.hasNext();) {
		    SkeletonHandle lodhandle=(SkeletonHandle)li.next();
		    if (lodhandle==skeleton.getLODPyramid().getDefaultLevel())
			continue;
		    // Write the motion out
		    writeMotion(lodhandle,frame,writer,false);		    
		}
	    }
	    writer.write("\n");
	}
    }
    
    /**
     * Write out the motion for a skeleton
     *
     * @param skeleton The skeleton.
     * @param frame The frame.
     * @param writer The writer.
     * @param first If this is the first skeleton.
     * @exception IOException if an error occurs
     */
    private void writeMotion(SkeletonHandle skeleton,int frame,Writer writer, boolean first) throws IOException
    {
	// Get the root joint
	Joint root=skeleton.getRootJoint();
	// Get the data manager
	DataManager dataManager=skeleton.getDataManager();
	if (first) {
	    // Write out the root's values - first translation
	    Matrix4d matrix=dataManager.getTranslationMatrix(root,frame);
	    Vector3d tv=new Vector3d();
	    matrix.get(tv);
	    writer.write(" "+tv.x+" "+tv.y+" "+tv.z);
	    // Now get the euler angles
	    Tuple3d rt=getRotation(dataManager,root,frame);
	    writer.write(" "+rt.z+" "+rt.x+" "+rt.y);
	}
	// Now write the channel
	writer.write(" "+root.getId());
	// Now write the elevation
	writer.write(" "+editor.getElevationFor(skeleton));
	// Iterate over the children
	for(Iterator i=root.getChildren().iterator();i.hasNext();) {
	    // 
	    Joint childJoint=(Joint)i.next();
	    // Recurse
	    writeJointMotion(dataManager,childJoint,frame,writer,first);
	}
    }
    
    /**
     * Get the rotation angles for a joint at a given frame.
     *
     * @param dataManager The data manager.
     * @param joint The joint.
     * @param frame The frame.
     * @return The rotation tuple.
     */
    private Tuple3d getRotation(DataManager dataManager,Joint joint,int frame)
    {
	// Get the rotation driver
	EulerRotationDriver driver=(EulerRotationDriver)dataManager.getRotationDriver(frame);
	return new Vector3d(driver.getRotateX(joint,frame)*180/Math.PI,
			    driver.getRotateY(joint,frame)*180/Math.PI,
			    driver.getRotateZ(joint,frame)*180/Math.PI);
    }
    
    /**
     * Write out the motion of one joint, and recurse to its children
     *
     * @param dataManager The data manager.
     * @param joint The joint.
     * @param frame The frame.
     * @param writer The writer.
     * @param first If this belongs to the first skeleton.
     * @exception IOException if an error occurs.
     */
    private void writeJointMotion(DataManager dataManager,Joint joint,int frame,Writer writer, boolean first) throws IOException
    {
	// Check for end of site
	if (joint.getChildren().size()==0) {
	    writer.write(" "+joint.getId());
	    return;
	}
	if (first) {
	    // Write out the channels
	    Tuple3d rt=getRotation(dataManager,joint,frame);
	    writer.write(" "+rt.z+" "+rt.x+ " "+rt.y);
	}
	
	writer.write(" "+joint.getId());
	// Iterate over the children
	for(Iterator i=joint.getChildren().iterator();i.hasNext();) {
	    Joint childJoint=(Joint)i.next();
	    // Recurse
	    writeJointMotion(dataManager,childJoint,frame,writer,first);
	}
    }
    
    /**
     * Export the model to an output stream.
     *
     * @param model The model
     * @param writer The output stream.
     */
    private void export(BasicAnimationModel model,Writer writer)
    {
	try {
	    // Write the hierarchy
	    writer.write("HIERARCHY\n");
	    exportSkeletons(model,writer);
	    // Write the channels
	    writer.write("MOTION\n");
	    exportMotion(model,writer);
	} 
	catch(IOException e) {
	    JOptionPane.showMessageDialog(this,"Error exporting: "+e.getMessage(),"Error",JOptionPane.ERROR_MESSAGE);
	}
    }

    /**
     * Export the skeleton hierarchy
     *
     * @param model The model
     * @param writer The writer
     * @exception IOException if an error occurs
     */
    private void exportSkeletons(BasicAnimationModel model,Writer writer) throws IOException
    {
	boolean isFirst=true;
	// Iterate over the handles
	for(Iterator hi=model.getHandles().iterator();hi.hasNext();) {
	    SkeletonHandle handle=(SkeletonHandle)hi.next();
	    exportSkeleton(handle,writer,true);
	    for(Iterator li=handle.getLODPyramid().getLevels();li.hasNext();) {
		SkeletonHandle lodhandle=(SkeletonHandle)li.next();
		if (lodhandle==handle.getLODPyramid().getDefaultLevel())
		    continue;
		// Export the skeleton
		exportSkeleton(lodhandle,writer,false);
	    }
	}
    }
    
    /**
     * Export one skeleton.
     *
     * @param handle The skeleton.
     * @param writer The writer.
     * @param first If this is the first skeleton
     * @exception IOException if an error occurs
     */
    private void exportSkeleton(SkeletonHandle handle,Writer writer,boolean first) throws IOException
    {
	// Get the root joint
	Joint root=handle.getRootJoint();
	// Write out the root
	writer.write("ROOT ");
	writer.write(root.getName()+"\n{\n");
	// Write out the offset
	tabs(1,writer);
	writer.write("OFFSET 0 0 0\n");
	// Write out the Channels
	tabs(1,writer);
	if (first) {
	    writer.write("CHANNELS 8 Xposition Yposition Zposition Zrotation Xrotation Yrotation Id Elevation\n");
	} else {
	    writer.write("CHANNELS 2 Id Elevation\n");
	}
	
	// Now write the joints
	for(Iterator i=root.getChildren().iterator();i.hasNext();) {
	    // Get the joint
	    Joint joint=(Joint)i.next();
	    // Write it out
	    exportJoint(handle,joint,writer,1,first);
	}
	writer.write("}\n");
    }
    
    /**
     * Export a joint.
     *
     * @param handle The handle.
     * @param joint The joint to export.
     * @param writer The output stream.
     * @param depth The depth.
     * @param first true if this is the first skeleton
     * @exception IOException if an error occurs
     */
    private void exportJoint(SkeletonHandle handle,Joint joint,Writer writer,int depth,boolean first) throws IOException
    {
	tabs(depth,writer);
	if (joint.getChildren().size()==0) {
	    writer.write("End Site\n");
	} else {
	    writer.write("JOINT ");
	    writer.write(joint.getName()+"\n");
	}
	tabs(depth,writer);
	writer.write("{\n");
	tabs(depth+1,writer);
	// Write the offset
	writer.write("OFFSET ");
	// Get the translation matrix
	Matrix4d tm=handle.getDataManager().getOffset(joint);
	// Get the vector
	Vector3d tv=new Vector3d();
	tm.get(tv);
	// Write it out
	writer.write(Double.toString(tv.x)+" "+Double.toString(tv.y)+" "+Double.toString(tv.z)+"\n");
	// Check for end site
	if (joint.getChildren().size()==0) {
	    tabs(depth+1,writer);
	    writer.write("CHANNELS 1 Id\n");
	    tabs(depth,writer);
	    writer.write("}\n");
	    return;
	}
	// Write the channels
	tabs(depth+1,writer);	
	if (first) {
	    writer.write("CHANNELS 4 Zrotation Xrotation Yrotation Id\n");
	} else {
	    writer.write("CHANNELS 1 Id\n");
	}
	
	// Iterate over the children
	for(Iterator i=joint.getChildren().iterator();i.hasNext();) {
	    Joint childJoint=(Joint)i.next();
	    // Recurse
	    exportJoint(handle,childJoint,writer,depth+1,first);
	}
	// Write the closing brace
	tabs(depth,writer);
	writer.write("}\n");
    }
    
    /**
     * Write out tabs.
     *
     * @param count The number of tabs
     * @param writer The writer
     * @exception IOException if an error occurs
     */
    private void tabs(int count,Writer writer) throws IOException
    {
	for(int i=0;i<count;i++) {
	    writer.write("  ");
	}
    }
    
    public static void main(String[] args) throws FileNotFoundException
    {
	GLModelView view=new GLModelView(null);
	final SkeletonEditorFrame frame=new SkeletonEditorFrame(view);
	frame.addWindowListener(new MyWindowAdapter());

	if (args.length>0) {
	    final File file=new File(args[0]);
	    SwingUtilities.invokeLater(new Runnable()
				   {
				       public void run()
					   {
					       frame.loadFile(file);
					   }
				   }
				   );
	}
    }
}

