package mc.marker;

import mc.io.bvh.ParseException;
import com.sun.java.util.collections.*;
import mc.model.AnimationModel;
import mc.model.ModelAdapter;
import mc.model.DataManager;
import mc.model.ModelEvent;
import mc.model.Joint;

import javax.vecmath.Matrix4d;
import javax.vecmath.Point3d;

import java.io.*;

/**
 * The class that handles the import and export of markers
 */
final public class MarkerManager extends ModelAdapter
{
    /** The animation model we are exporting */
    private AnimationModel model;
    
    /**
     * The markers
     */
    private final Vector markers;

    /** The number of millis that the thread will sleep*/
    private final long timeout;
    
    /**
     * The matrix cache. Indexed by joints
     */
    private final Map matrixCache;
    
    /**
     * The joint map. Indexed by joint names
     */
    private final Map jointMap;
    
    /** The writer that will be used to spit out the marker information */
    private Writer writer;

    /**
     * The timestamp for the start of the export
     */
    private long millis;
    
    /**
     * Initialize the manager from a file.
     *
     * @param s The input stream that contains the marker information
     * @param timeout The interval between frames in milliseconds.
     * @exception ParseException if a parse error occurs
     */
    public MarkerManager(InputStream s,long timeout) throws ParseException
    {
	// Initialize the variables of the class
	this.markers=new Vector();
	this.matrixCache=new HashMap();
	this.jointMap=new HashMap();
	this.timeout=timeout;
	
	// Read in the stream
	readStream(s);
    }

    /**
     * Build the name to joint map.
     *
     * @param model The model
     */
    private void buildJointMap(AnimationModel model)
    {
	jointMap.clear();
	// Get the root joint
	recordName(model.getFirstSkeleton().getRootJoint());
    }
    
    /**
     * Recursively descent the joint tree and record the
     * name of each joint
     *
     * @param joint The joint.
     */
    private void recordName(Joint joint)
    {
	jointMap.put(joint.getName(),joint);
	for(Iterator i=joint.getChildren().iterator();i.hasNext();) {
	    Joint childJoint=(Joint)i.next();
	    // Recurse
	    recordName(childJoint);
	}
    }
    
    /**
     * The model was modified
     *
     * @param e The model event.
     */
    public void frameTicked(int currentFrame,int increment) 
    {
	// Write out the current frame and current time
	try {
	    writer.write(Integer.toString(model.getCurrentFrame()));
	    writer.write(","+getTime());
	}
	catch(IOException e) {
	    throw new Error("Unexpected: "+e);
	}
	
	// Traverse the markers
	for(Iterator i=markers.iterator();i.hasNext();) {
	    // Get the marker info
	    MarkerInfo markerInfo=(MarkerInfo)i.next();
	    // Get the joint that the marker is attached to
	    Joint joint=(Joint)jointMap.get(markerInfo.jointName);
	    // Get the matrix of the joint
	    Matrix4d matrix=getMatrix(joint);
	    // Multiply the matrix with the translation of the marker
	    Point3d resultPoint=new Point3d();
	    matrix.transform(markerInfo.translation,resultPoint);
	    // Write out the result point
	    try {
		writer.write(","+resultPoint.x+","+resultPoint.y+","+resultPoint.z);
	    }
	    catch(IOException e) {
		throw new Error("Unexpected: "+e);
	    }
	}
	// Write the end of line
	try {
	    writer.write("\n");
	    System.err.println("Written out frame: "+model.getCurrentFrame());
	}
	catch(IOException e) {
	    throw new Error("Unexpected: "+e);
	}
    }
    
    /**
     * Write out the header of the marker file
     *
     */
    private void writeHeader()
    {
	try {
	    writer.write("KSample, KTime, ");
	    for(Iterator i=markers.iterator();i.hasNext();) {
		MarkerInfo markerInfo=(MarkerInfo)i.next();
		// Write out the channel names
		writer.write(markerInfo.name+"-X,"+markerInfo.name+"-Y,"+markerInfo.name+"-Z");
	    }
	    // End of line
	    writer.write("\n");
	}
	catch(IOException e) {
	    throw new Error("Unexpected: "+e);
	}
    }
    
    /**
     * Output the marker information from an animation model.
     *
     * @param model The animation model.
     * @param out The output stream.
     */
    public void export(AnimationModel model,OutputStream out)
    {
	this.model=model;
	// Create the writer
	this.writer=new BufferedWriter(new OutputStreamWriter(out));
	// Initialize the joint map
	buildJointMap(model);
	// Clear the cache
	matrixCache.clear();
	// Output the header
	writeHeader();
	// Start ticking the clock
	model.addModelListener(this);
	// Record the time stamp
	millis=System.currentTimeMillis();
	// Play the first tick
	model.tick(1);
	// Replay now
	while(model.getCurrentFrame()!=model.getNumberOfFrames()-1) {
	    // Sleep
	    try {
		Thread.sleep(timeout);
	    }
	    catch(InterruptedException e) {
		throw new Error("Unexpected "+e);
	    }
	    // Clear the cache
	    matrixCache.clear();
	    // Tick one
	    model.tick(1);
	}
	// Close the writer
	try {
	    writer.flush();
	    writer.close();
	}
	catch(IOException e) {
	    System.err.println("Unexpected: "+e.getMessage());
	}
    }
    
    /**
     * Get back a time stamp
     *
     * @return The time stamp.
     */
    private long getTime()
    {
	return System.currentTimeMillis()-millis;
    }
    
    /**
     * Get the matrix of a joint. First, we look in the cache.
     * If we have a cache miss, we recursively get the matrix of
     * the parent nodes, and multiply everything to form the new matrix.
     *
     * @param joint The joint.
     * @return The 4x4 matrix of the node.
     */
    private Matrix4d getMatrix(Joint joint)
    {
	// Look in the cache
	Matrix4d matrix=(Matrix4d)matrixCache.get(joint);
	if (matrix!=null) {
	    // System.err.println("Cached matrix\n "+matrix+ " for "+joint.getName());
	    return matrix;
	}
	
	DataManager manager=model.getFirstSkeleton().getDataManager();
	// Check if this is the root
	if (joint.getParent()==null) {
	    // Record the matrices
	    matrix=new Matrix4d(getOrthoMatrix(joint));
	    matrixCache.put(joint,matrix);
	    // System.err.println("Caching matrix\n"+matrix+" for "+joint.getName());
	    return matrix;
	}
	
	// Get the parent's matrix
	matrix=new Matrix4d(getMatrix(joint.getParent()));
	matrix.mul(getOrthoMatrix(joint));
	// Record it
	matrixCache.put(joint,matrix);
	// System.err.println("Caching matrix\n"+matrix+" for "+joint.getName());
	return matrix;
    }
    
    /**
     * Return the matrix of a joint. This is the translation matrix multiplied
     * by the rotation matrix.
     *
     * @param joint The joint.
     */
    private Matrix4d getOrthoMatrix(Joint joint)
    {
	Matrix4d result=new Matrix4d(model.getFirstSkeleton().getDataManager().getTranslationMatrix(joint));
	result.mul(model.getFirstSkeleton().getDataManager().getRotationMatrix(joint));
	return result;
    }
    
    /**
     * Parse a stream of marker information
     *
     * @param s The input stream
     * @exception ParseException if a parse error occurs
     */
    private void readStream(InputStream s) throws ParseException
    {
	// Start reading lines from the stream
	LineNumberReader reader=new LineNumberReader(new InputStreamReader(s));
	String line=null;
	
	try {
	    while((line=reader.readLine())!=null) {
		// Parse the line read
		if (parseLine(line)==true) {
		    return;
		}
	    }
	}
	catch(IOException e) {
	    throw new ParseException("IO Exception: "+e);
	}
	
	throw new ParseException("End of file seen before END marker");
    }
    
    /**
     * Parse one line of input
     *
     * @param line The line
     * @return true if the END token has been detected
     * @exception ParseException if a parse error occurs
     */
    private boolean parseLine(String line) throws ParseException
    {
	StreamTokenizer tokenizer=new StreamTokenizer(new StringReader(line));
	tokenizer.eolIsSignificant(true);
	
	// The name of the marker
	String name=null;
	// The name of the joint
	String joint=null;
	// The x y z variables
	double x=0;
	double y=0;
	double z=0;
	int count=0;
	
	while(true) {
	    int token;
	    
	    try {
		token=tokenizer.nextToken();
	    }
	    catch(IOException e) {
		throw new ParseException(e.getMessage());
	    }
	    
	    switch(token) {
	    case StreamTokenizer.TT_EOL:
	    case StreamTokenizer.TT_EOF:
		// check that we have everything we need
		if (name==null || joint==null || count!=3) {
		    throw new ParseException("Could not read info from "+line);
		}
		markers.add(new MarkerInfo(name,joint,x,y,z));
		return false;
	    case StreamTokenizer.TT_WORD:
		if (name==null && tokenizer.sval.equalsIgnoreCase("end")) {
		    return true;
		}
		
		if (name==null) {
		    name=tokenizer.sval;
		} else if (joint==null) {
		    joint=tokenizer.sval;
		} else {
		    throw new ParseException("Unexpected string: "+tokenizer.sval);
		}
		break;
	    case StreamTokenizer.TT_NUMBER:
		switch(count) {
		case 0:
		    x=tokenizer.nval;
		    break;
		case 1:
		    y=tokenizer.nval;
		    break;
		case 2:
		    z=tokenizer.nval;
		    break;
		default:
		    throw new ParseException("Too many numbers!");
		}
		count++;
		break;
	    default:
		throw new ParseException("Unknwon token "+token);
	    }
	}
    }
}

		
		
