package mc.model;

import com.sun.java.util.collections.*;

/**
 * The animation model interface.
 */
public interface  AnimationModel
{
    /**
     * Get the current frame.
     *
     * @return The current frame.
     */
    public int getCurrentFrame();
    
    /**
     * Get the number of frames.
     *
     * @return The number of frames.
     */
    public int getNumberOfFrames();
    
    /**
     * Get a list of the handles of the
     * skeletons.
     *
     * @return A list of the handles.
     */
    public Collection getHandles();

    /**
     * Get the first skeleton handle. 
     * Useful for backewards compatibility when there was
     * a one model assumption.
     * @return skeleton handle
     */
    public SkeletonHandle getFirstSkeleton();

    /**
     * Add a model listener
     *
     * @param l The model listener.
     */
    public void addModelListener(ModelListener l);

    /**
     * Remove a model listener.
     *
     * @param l The model listener to remove.
     */
    public void removeModelListener(ModelListener l);

    /**
     * Cause time to advance. The time is rounded to the start and end
     * frames if the increment causes time to exceed the range of frames. After time
     * is advanced, the joint hierarchy is refreshed, and the listeners notified.
     *
     * @param increment The increment in time. 
     */
    public void tick(int increment);

    /**
     * Automatically play the frames and refresh the model
     *
     * @param increment The increment for advancing frames.
     */
    public void play(int increment);
    
    /**
     * Set the behavior of the model when the end or start
     * frame are reached. If the parameter is true, then
     * frames are wrapped around.
     *
     * @param loop The looping behavior.
     */
    public void setLoop(boolean loop);
    
    /**
     * Get the looping behavior of the model.
     *
     * @return true/false
     */
    public boolean isLooping();
    
    /**
     * stop the animation.
     *
     */
    public void stop();
    
    /**
     * Directly set the frame of the model.
     *
     * @param frame A frame from 0 to getNumberOfFrames()-1.
     */
    public void setFrame(int frame);

    /**
     * Get the driver factory associated with this animation model.
     *
     * @return The driver factory.
     */
    public DriverFactory getDriverFactory();

    /**
     * Get the root joint.
     * 
     * @return The root joint
     * @deprecated Use the {@link #getHandles()} method.
     */
    //public Joint getRootJoint();
    
    /**
     * Get the data manager
     *
     * @return a value of type 'DataManager'
     * @deprecated Use  {@link #getHandles()} method to get the data manager 
     * for a skeleton
     */
    //public DataManager getDataManager();
}
