package mc.model;

import com.sun.java.util.collections.*;
import java.awt.event.*;

/**
 * The basic implementation of the animation model.
 */
public class BasicAnimationModel implements AnimationModel, TimerListener
{
    /**
     * The model listeners.
     */
    private List modelListeners;
    
    /** The number of frames */
    private int numberOfFrames;

    /** The timer */
    private final ModelTimer timer;
    
    /**
     * The handles list
     */
    private final Set handles;
    
    /** 
     * The data manager for the first skeleton 
     * @deprecated 
     */
    private DataManager theDataManager;
    
    /** 
     * The root joint for the first skeleton 
     * @deprecated
     */
    private Joint rootJoint;
    
    /**
     * The current frame.
     */
    private int currentFrame;

    /** the looping behavior */
    private boolean loop;
    
    /** The driver factory */
    private DriverFactory theDriverFactory;
    
    /**
     * Initialize the model.
     *
     */
    public BasicAnimationModel(ModelTimer timer)
    {
	this.numberOfFrames=0;
	this.currentFrame=0;
	this.handles=new TreeSet();
	this.modelListeners=new LinkedList();
	this.timer=timer;
	this.timer.addListener(this);
	this.loop=false;
	this.theDataManager=null;
	this.rootJoint=null;
    }
    
    /**
     * Set the driver factory.
     *
     * @param theDriverFactory The driver factory.
     */
    public void setDriverFactory(DriverFactory theDriverFactory)
    {
	this.theDriverFactory=theDriverFactory;
    }
    
    public DriverFactory getDriverFactory()
    {
	return theDriverFactory;
    }
    
    /**
     * Initialize the given model with a number of frames.
     *
     * @param numberOfFrames The number of frames.
     */
    public void setNumberOfFrames(int numberOfFrames)
    {
	this.numberOfFrames=numberOfFrames;
    }
    
    /**
     * Return the number of frames.
     *
     * @return The number of frames.
     */
    public int getNumberOfFrames()
    {
	return numberOfFrames;
    }

    /**
     * Add a skeleton handle to the model, by cloning
     * the data manager of a given handle.
     *
     * @param skeleton The skeleton to add.
     * @return The skeleton handle
     */
    public SkeletonHandle addSkeleton(SkeletonHandle skeleton)
    {
	DataManager dm=new DataManager();
	dm.copyOffsetsFrom(skeleton.getDataManager());
	dm.copyDriversFrom(skeleton.getDataManager());
	dm.setAnimationModel(this);
	
	SkeletonHandle handle=new SkeletonHandle(skeleton.getRootJoint(),dm);
	handles.add(handle);
	return handle;
    }
    
    /**
     * Add a new skeleton to this animatin model.
     *
     * @param joint The root joint of the skeleton.
     * @return The handle
     */
    public SkeletonHandle addSkeleton(Joint joint)
    {
	DataManager dm=new DataManager();
	dm.setAnimationModel(this);

	SkeletonHandle handle=new SkeletonHandle(joint,dm);
	handles.add(handle);

	// Cache the first skeleton
	if (rootJoint==null) {
	    rootJoint=joint;
	    theDataManager=dm;
	} 
	    
	// Fire a notification
	fireNotification(new ModelEvent(this,ModelEvent.SKELETON_ADDED,handle));
	
	return handle;
    }

    /**
     * Create a new skeleton thats suitable for this AM
     *
     * @param joint The root joint of the skeleton.
     * @return The handle
     */
    public SkeletonHandle createSkeleton(Joint joint)
    {
	DataManager dm=new DataManager();
	dm.setAnimationModel(this);

	SkeletonHandle handle=new SkeletonHandle(joint,dm);
	
	return handle;
    }

    /**
     * Return the handles to the skeletons.
     *
     * @return The handles to the skeletons.
     */
    public Collection getHandles()
    {
	return handles;
    }

    public SkeletonHandle getFirstSkeleton()
    {
	Iterator i = handles.iterator();
	if(i.hasNext()) 
	    return (SkeletonHandle)i.next();
	return null;
    }

    /**
     * Replace the root joint of a skeleton.
     *
     * @param skeleton The skeleton handle.
     * @param newRoot The new joint.
     */
    public void setJoint(SkeletonHandle skeleton,Joint newRoot)
    {
	skeleton.setRootJoint(newRoot);
	fireNotification(new ModelEvent(this,ModelEvent.SKELETON_CHANGED,skeleton));
    }

    /**
     * Get the current frame number.
     *
     * @return The current frame.
     */
    public int getCurrentFrame()
    {
	return currentFrame;
    }

    public void timerFired(TimerEvent e)
    {
	if (e.type==ModelTimer.TICK) {
	    timerExpired(e.increment);
	} 
	else if (e.type==ModelTimer.STOP) {
	    fireNotification(new ModelEvent(this,ModelEvent.ANIMATION_STOPPED));
	}
    }

    /**
     * Stop the timer.
     *
     */
    public void stop()
    {
	stopAnimation();
    }
    
    private void stopAnimation()
    {
	if (timer.isRunning()) {
	    timer.stop();
	}
    }
    
    /**
     * Add a model listener.
     *
     * @param l The model listener.
     */
    public void addModelListener(ModelListener l)
    {
	modelListeners.add(l);
    }
    
    /**
     * Remove a model listener.
     *
     * @param l The model listener.
     */
    public void removeModelListener(ModelListener l)
    {
	modelListeners.remove(l);
    }
    
    /**
     * Notify the listeners that the model has been modified.
     *
     */
    public void fireNotification(ModelEvent event)
    {
	// Iterate over the listeners
	for(Iterator i=modelListeners.iterator();i.hasNext();) {
	    ModelListener listener=(ModelListener)i.next();
	    // Notify
	    listener.modelModified(event);
	}
    }

    /**
     * Tick!
     *
     * @param increment The increment in frames
     */
    public void tick(int increment)
    {
	if (increment==0)
	    return;
	
	// Set the frame
	setFrame(currentFrame+increment);
    }

    /**
     * Set the current frame and notify the listeners.
     *
     * @param frameNumber The frame.
     */
    private void setFrameInternal(int frameNumber)
    {
	// Save the current frame
	int oldFrame=currentFrame;
	
	if (loop) {
	    currentFrame=(frameNumber+numberOfFrames)%numberOfFrames;
	} else {
	    currentFrame=Math.max(0,frameNumber);
	    currentFrame=Math.min(currentFrame,numberOfFrames-1);
	}
	
	// Check for stop of animation
	if (oldFrame==currentFrame && timer.isRunning()) {
	    stopAnimation();
	    return;
	} else {
	    notifyTick(currentFrame,currentFrame-oldFrame);
	}
    }

    /**
     * Notify the listeners that a frame tick has occured.
     *
     * @param currentFrame The current frame.
     * @param increment The increment.
     */
    protected void notifyTick(int currentFrame,int increment)
    {
	// Notify the listeners
	for(Iterator li=modelListeners.iterator();li.hasNext();) {
	    ModelListener l=(ModelListener)li.next();
	    l.frameTicked(currentFrame,increment);
	}
    }
    
    /**
     * Set the current frame. If the timer is
     * playing, stop it.
     *
     * @param frameNumber The frame number.
     */
    public void setFrame(int frameNumber)
    {
	// Stop the timer if it is running
	stopAnimation();
	// 
	setFrameInternal(frameNumber);
    }
    
    /**
     * Notification that the timer has expired
     *
     */
    private void timerExpired(int playIncrement)
    {
	// Set the current frame 
	setFrameInternal(currentFrame+playIncrement);
    }

    public boolean isLooping()
    {
	return loop;
    }
    
    public void setLoop(boolean loop)
    {
	this.loop=loop;
	fireNotification(new ModelEvent(this,ModelEvent.LOOP_CHANGED));
    }

    public void play(int increment)
    {
	// Check if we should stop the animation
	stopAnimation();
	// Check if the timer has started
	if (!timer.isRunning()) {
	    timer.start(increment);
	}
    }
}
