package mc.model;

import com.sun.java.util.collections.*;
import javax.vecmath.Vector3d;
import javax.vecmath.Matrix4d;

/**
 * The basic driver that uses precomputed data to drive
 * the whole process.
 */
public class BasicTranslationDriver implements MutableTranslationDriver
{
    /** The map for translation values */
    private final Map translationValuesMap;

    /** The translation matrix map */
    private Map translationMatrixMap;
    
    /** The number of frames */
    private int numberOfFrames;
    
    static private final int TRANSLATEX=0;
    
    static private final int TRANSLATEY=1;
    
    static private final int TRANSLATEZ=2;
    
    /**
     * Initialize.
     *
     */
    public BasicTranslationDriver()
    {
	this.translationValuesMap=new HashMap();
	this.translationMatrixMap=new HashMap();
    }
    
    /**
     * Initialize the driver for a number
     * frames.
     *
     * @param numberOfFrames The number of frames
     */
    public void initialize(int numberOfFrames)
    {
	translationValuesMap.clear();
	this.numberOfFrames=numberOfFrames;
    }
    
    /**
     * Get the array for a joint. If the joint is not
     * associated with an array, one is created and registered.
     *
     * @param joint The joint.
     * @param map The map of joints->arrays
     * @return The array of the joint.
     */
    private FrameDataArray getArrayFor(Joint joint,Map map)
    {
	// Index in the map
	FrameDataArray array=(FrameDataArray)map.get(joint);
	if (array==null) {
	    array=new FrameDataArray(numberOfFrames);
	    map.put(joint,array);
	}
	return array;
    }

    /**
     * Worker method. Set the value in the array of a node, for
     * a specific frame number.
     *
     * @param joint The joint.
     * @param frameNumber The frame number
     * @param value The value.
     * @param arrayMap The map of arrrays
     * @param index The index of the value
     */
    private void setValue(Joint joint,int frameNumber,double value,Map arrayMap,int index)
    {
	// Look up the frame data for joint
	FrameDataArray frameData=getArrayFor(joint,arrayMap);
	// Set the appropriate value
	frameData.setValue(frameNumber,index,value);
    }

    /**
     * Worker method. Get the value of the array of 
     * a joint for the current frame.
     *
     * @param joint The joint.
     * @param frameNumber The frame number.
     * @param arrayMap The array map.
     * @param index The index.
     * @return the value.
     */
    private double getValue(Joint joint,int frameNumber,Map arrayMap,int index)
    {
	// Get the map
	FrameDataArray frameData=getArrayFor(joint,arrayMap);
	// Return the value
	return frameData.getValue(frameNumber,index);
    }
    
    /**
     * Set the translateX for a joint.
     *
     * @param joint The joint.
     * @param frameNumber The frame number.
     * @param value The value.
     */
    public void setTranslateX(Joint joint,int frameNumber,double value)
    {
	setValue(joint,frameNumber,value,translationValuesMap,TRANSLATEX);
	// Update the matrix
	Matrix4d[] matrices=getTranslationMatrices(joint);
	Matrix4d xlat=new Matrix4d();
	xlat.setIdentity();
	xlat.set(new Vector3d(value,0,0));
	matrices[frameNumber].mul(xlat);
    }

    /**
     * Set the translateY for a joint.
     *
     * @param joint The joint.
     * @param frameNumber The frame number.
     * @param value The value.
     */
    public void setTranslateY(Joint joint,int frameNumber,double value)
    {
	setValue(joint,frameNumber,value,translationValuesMap,TRANSLATEY);
	// Update the matrix
	Matrix4d[] matrices=getTranslationMatrices(joint);
	Matrix4d xlat=new Matrix4d();
	xlat.setIdentity();
	xlat.set(new Vector3d(0,value,0));
	matrices[frameNumber].mul(xlat);
    }


    /**
     * Set the translateZ for a joint.
     *
     * @param joint The joint.
     * @param frameNumber The frame number.
     * @param value The value.
     */
    public void setTranslateZ(Joint joint,int frameNumber,double value)
    {
	setValue(joint,frameNumber,value,translationValuesMap,TRANSLATEZ);
	// Update the matrix
	Matrix4d[] matrices=getTranslationMatrices(joint);
	Matrix4d xlat=new Matrix4d();
	xlat.setIdentity();
	xlat.set(new Vector3d(0,0,value));
	matrices[frameNumber].mul(xlat);
    }

    /**
     * Get the translate X for a joint.
     *
     * @param joint The joint.
     * @param frameNumber The frame number
     * @return the value.
     */
    public double getTranslateX(Joint joint,int frameNumber)
    {
	return getValue(joint,frameNumber,translationValuesMap,TRANSLATEX);
    }

    /**
     * Get the translate Y for a joint.
     *
     * @param joint The joint.
     * @param frameNumber The frame number.
     * @return the value.
     */
    public double getTranslateY(Joint joint,int frameNumber)
    {
	return getValue(joint,frameNumber,translationValuesMap,TRANSLATEY);
    }

    /**
     * Get the translate Z for a joint.
     *
     * @param joint The joint.
     * @param frameNumber the frame number.
     * @return the value.
     */
    public double getTranslateZ(Joint joint,int frameNumber)
    {
	return getValue(joint,frameNumber,translationValuesMap,TRANSLATEZ);
    }

    /**
     * Get the translation matrix of a joint for a given frame.
     *
     * @param joint The joint
     * @param frame The frame
     * @return The 4x4 translation matrix
     */
    public Matrix4d getTranslationMatrix(Joint joint,int frame)
    {
	Matrix4d[] matrices=getTranslationMatrices(joint);
	return matrices[frame];
    }
    
    /**
     * Return the 3-vector representation of a rotation.
     *
     * @param joint The joint.
     * @param frame The frame.
     * @return The RotationVector3d
     */
    public DataRepresentation getTranslationRep(Joint joint,int frame)
    {
	return new TranslationVector3d(getTranslateX(joint,frame),getTranslateY(joint,frame),getTranslateZ(joint,frame));
    }

    /**
     * Get the translation matrices for a specific joint. If the
     * matrices do not exist, a new array is created where all matrices
     * are set to the identity matrix.
     *
     * @param joint The joint.
     * @return The array of 4x4 matrices.
     */
    private Matrix4d[] getTranslationMatrices(Joint joint)
    {
	return getMatrices(translationMatrixMap,joint);
    }
    
    /**
     * Worker method. Check if the map contains the matrices for
     * a specific joint. If not, the array of matrices is created
     * and all matrices assigned the identity matrix.
     *
     * @param matricesMap The map to search.
     * @param joint The joint.
     * @return The array of 4x4 Matrices for the joint.
     */
    private Matrix4d[] getMatrices(Map matricesMap,Joint joint)
    {
	// Index in the map
	Matrix4d[] matrices=(Matrix4d[])matricesMap.get(joint);
	if (matrices==null) {
	    matrices=new Matrix4d[numberOfFrames];
	    for(int i=0;i<numberOfFrames;i++) {
		matrices[i]=new Matrix4d();
		matrices[i].setIdentity();
	    }

	    // Register it
	    matricesMap.put(joint,matrices);
	}
	return matrices;
    }
    
}
