package mc.model;

import javax.vecmath.Matrix4d;
import com.sun.java.util.collections.*;
import javax.vecmath.Vector3d;

/**
 * The class that feeds joints with their values.
 */
public class DataManager implements DataSource
{
    /**
     * The list of rotation drivers.
     */
    private List rotationDrivers;
    
    /**
     * The list of translation drivers.
     */
    private List translationDrivers;

    /**
     * The animation model.
     */
    private AnimationModel animationModel;

    /** The listeners */
    private List listeners;

    /** The map with the offsets */
    private Map offsets;
    
    /**
     * Initialize the driver manager.
     *
     */
    public DataManager()
    {
	this.rotationDrivers=new LinkedList();
	this.translationDrivers=new LinkedList();
	this.animationModel=null;
	this.listeners=new LinkedList();
	this.offsets=new HashMap();
	
    }
    
    /**
     * Set the animation model of the data manager.
     *
     * @param theAnimationModel The animation model.
     */
    public void setAnimationModel(AnimationModel theAnimationModel)
    {
	this.animationModel=theAnimationModel;
    }

    /**
     * Get the rotation driver that is responsible for
     * a given frame.
     *
     * @param currentFrame The frame.
     * @return The rotation driver.
     */
    public RotationDriver getRotationDriver(int currentFrame)
    {
	// Search the list
	DriverData driverData=getDriverData(rotationDrivers,currentFrame);
	return (RotationDriver)driverData.driver;
    }
    
    /**
     * Set the offset data for a joint.
     *
     * @param joint The joint.
     * @param x The x.
     * @param y The y.
     * @param z The z.
     */
    public void setOffset(Joint joint,double x,double y,double z)
    {
	Matrix4d offset=new Matrix4d();
	offset.set(new Vector3d(x,y,z));
	offsets.put(joint,offset);
    }
    
    public void setOffset(Joint joint,Matrix4d offset)
    {
	offsets.put(joint,offset);
    }
    
    public void dumpOffsets()
    {
	for(Iterator i=offsets.keySet().iterator();i.hasNext();) {
	    Joint joint=(Joint)i.next();
	    System.err.print("Joint; "+joint.getName());
	    System.err.print(" Matrix:");
	    
	    Matrix4d matrix=(Matrix4d)offsets.get(joint);
	    if (matrix==null) {
		System.err.println("#ERROR");
	    } else {
		System.err.println(matrix);
	    }
	}
    }
    
    /**
     * Get the offset matrix of a node.
     *
     * @param joint The joint.
     * @return The matrix.
     */
    public Matrix4d getOffset(Joint joint)
    {
	Matrix4d result=(Matrix4d)offsets.get(joint);

	if (result==null){
	    throw new Error("Cannot find offset for joint: "+joint.getName());
	}
	return result;
    }
    
    /**
     * Get the xlat driver that is responsible for a given frame
     * number.
     *
     * @param currentFrame The current frame.
     * @return the translation driver
     */
    public TranslationDriver getTranslationDriver(int currentFrame)
    {
	DriverData driverData=getDriverData(translationDrivers,currentFrame);
	return (TranslationDriver)driverData.driver;
    }

    public void printData()
    {
	System.err.println("Translation:");
	for(Iterator i=translationDrivers.iterator();i.hasNext();) {
	    DriverData data=(DriverData)i.next();
	    System.err.println("From:"+data.startFrame+" To:"+data.endFrame+" Driver:"+data.driver);
	}
	System.err.println("Rotation:");
	for(Iterator i=rotationDrivers.iterator();i.hasNext();) {
	    DriverData data=(DriverData)i.next();
	    System.err.println("From:"+data.startFrame+" To:"+data.endFrame+" Driver:"+data.driver);
	}
    }
    
    /**
     * Search a list of drivers and return the driver
     * that is responsible for a given frame.
     *
     * @param driversList The list of drivers
     * @param frame The frame number
     * @return The driver data object
     */
    private DriverData getDriverData(List driversList,int frame)
    {
	for(Iterator li=driversList.iterator();li.hasNext();) {
	    DriverData driverInfo=(DriverData)li.next();
	    // Check if the driver qualifies
	    if (driverInfo.startFrame<=frame && frame<=driverInfo.endFrame) {
		return driverInfo;
	    }
	}
	return null;
    }

    /**
     * Add a rotation driver for a specified range of frames.
     *
     * @param startFrame The start frame.
     * @param endFrame The end frame.
     * @param rotDriver The channel driver to add.
     */
    public void addRotationDriver(int startFrame,int endFrame,RotationDriver rotDriver) throws IllegalArgumentException
    {
	rotationDrivers.add(0,new DriverData(rotDriver,startFrame,endFrame));
	fireNotification(new DataManagerEvent(this,DataManagerEvent.ROTATION_DRIVER_ADDED,startFrame,endFrame,rotDriver.getDescription()));
    }
    
    /**
     * Get the internal representation of the rotation data
     * for a joint and the current frame.
     *
     * @param joint The joint.
     * @return The internal representation.
     */
    public DataRepresentation getRotationRep(Joint joint)
    {
	//System.err.println(this+": Getting rotation for "+joint.getName());
	//System.err.println(rotationDrivers.size());
	
	RotationDriver driver=getRotationDriver(animationModel.getCurrentFrame());
	return driver.getRotationRep(joint,animationModel.getCurrentFrame());
    }
    
    /**
     * Get the internal representation of the translation data
     * for a joint and the current frame.
     *
     * @param joint The joint.
     * @return The internal representation.
     */
    public DataRepresentation getTranslationRep(Joint joint)
    {
	TranslationDriver driver=getTranslationDriver(animationModel.getCurrentFrame());
	return driver.getTranslationRep(joint,animationModel.getCurrentFrame());
    }
    
    /**
     * Add a rotation driver for a specified range of frames.
     *
     * @param startFrame The start frame.
     * @param endFrame The end frame.
     * @param xlatDriver The channel driver to add.
     * @exception IllegalArgumentException if the encoding of the different driver is not compatible
     */    
    public void addTranslationDriver(int startFrame,int endFrame,TranslationDriver xlatDriver) throws IllegalArgumentException
    {
	translationDrivers.add(0,new DriverData(xlatDriver,startFrame,endFrame));
	fireNotification(new DataManagerEvent(this,DataManagerEvent.TRANSLATION_DRIVER_ADDED,startFrame,endFrame,null));
    }

    /**
     * Convenience method. Adds a driver for the whole range 
     * of frames.
     *
     * @param driver The xlat driver.
     */
    public void addTranslationDriver(TranslationDriver driver)
    {
	addTranslationDriver(0,animationModel.getNumberOfFrames(),driver);
    }
    
    /**
     * Remvoe a translation driver for a specified frame range. The first
     * driver that handles the given range will be deleted.
     *
     * @param startFrame Teh start frame.
     * @param endFrame The end frame.
     */
    public void removeTranslationDriver(int startFrame,int endFrame)
    {
	removeDriver(translationDrivers,startFrame,endFrame);
	fireNotification(new DataManagerEvent(this,DataManagerEvent.TRANSLATION_DRIVER_REMOVED,startFrame,endFrame,null));
    }
    
    /**
     * Remvoe a rotation driver for a specified frame range. The first
     * driver that handles the given range will be deleted.
     *
     * @param startFrame Teh start frame.
     * @param endFrame The end frame.
     */
    public void removeRotationDriver(int startFrame,int endFrame)
    {
	removeDriver(rotationDrivers,startFrame,endFrame);
	fireNotification(new DataManagerEvent(this,DataManagerEvent.ROTATION_DRIVER_REMOVED,startFrame,endFrame,null));
    }

    /**
     * Worker method. Removes the first driver that handles a specified range 
     * of frames.
     *
     * @param drivers the vector of drivers.
     * @param startFrame the start frame.
     * @param endFrame the end frame.
     */
    private void removeDriver(List drivers,int startFrame,int endFrame)
    {
	// Iterate and check the range
	for(ListIterator i=drivers.listIterator();i.hasNext();) {
	    DriverData data=(DriverData)i.next();
	    // 
	    if (data.startFrame==startFrame && data.endFrame==endFrame) {
		i.remove();
		break;
	    }
	}
    }
    
    /**
     * Get the translation matrix for a joint for a specified
     * frame.
     *
     * @param joint The joint.
     * @param frameNumber The frame. The first frame is 0.
     * @return The 4x4 translation matrix.
     */
    public Matrix4d getTranslationMatrix(Joint joint,int frameNumber)
    {
	// Get the driver
	TranslationDriver driver=getTranslationDriver(frameNumber);
	// 
	Matrix4d translation=driver.getTranslationMatrix(joint,frameNumber);
	if (translation==null) {
	    throw new Error("Cannot find translation for "+joint.getName());
	}
	
	Matrix4d result=new Matrix4d();
	Matrix4d offset=(Matrix4d)offsets.get(joint);
	if (offset==null) {
	    throw new Error("Cannot find offset for "+joint.getName());
	}
	result.mul(offset,translation);
//    	if (joint.getParent()!=null  && joint.getParent().getName().equals("LeftUpArm")) {
//    	    System.err.println(joint+" Xlat: "+result);
//    	}
	return result;
    }

    /**
     * Delegate to getTranslationMatrix for the current frame.
     *
     * @param joint The joint.
     * @return The 4x4 translation matrix.
     */
    public Matrix4d getTranslationMatrix(Joint joint)
    {
	return getTranslationMatrix(joint,animationModel.getCurrentFrame());
    }
    
    /**
     * Get the rotation matrix for a joint for a specified
     * frame.
     *
     * @param joint The joint.
     * @param frameNumber The frame.
     * @return The 4x4 rotation matrix. 
     */
    public Matrix4d getRotationMatrix(Joint joint,int frameNumber)
    {
	// Get the driver
	//System.err.println(this+": Getting matrix for "+joint.getName());
	RotationDriver driver=getRotationDriver(frameNumber);
	return driver.getRotationMatrix(joint,frameNumber);
    }
    
    /**
     * Get the rotation matrix of the joint for the current frame.
     *
     * @param joint The joint.
     * @return The 4x4 rotation matrix.
     */
    public Matrix4d getRotationMatrix(Joint joint)
    {
	return getRotationMatrix(joint,animationModel.getCurrentFrame());
    }
    
    /**
     * Get the rotation driver that handles the range
     * of frames. 
     *
     * @param startFrame The start frame.
     * @param endFrame The end frame.
     * @return The rotation driver that handles the range, or null if none exists
     */
    public RotationDriver getRotationDriverFor(int startFrame,int endFrame)
    {
	return (RotationDriver)getDriverFor(rotationDrivers,startFrame,endFrame);
    }
    
    /**
     * Get the translation driver that handles the range
     * of frames. 
     *
     * @param startFrame The start frame.
     * @param endFrame The end frame.
     * @return The rotation driver that handles the range, or null if none exists
     */
    public TranslationDriver getTranslationDriverFor(int startFrame,int endFrame)
    {
	return (TranslationDriver)getDriverFor(translationDrivers,startFrame,endFrame);
    }

    /**
     * Worker method. Find the driver that handles a range of frames.
     *
     * @param drivers The list of drivers.
     * @param startFrame The start frame.
     * @param endFrame The end frame.
     * @return The driver object.
     */
    private Object getDriverFor(List drivers,int startFrame,int endFrame)
    {
	// Iterate over the list 
	for(Iterator i=drivers.iterator();i.hasNext();) {
	    // Get the data
	    DriverData data=(DriverData)i.next();
	    if (startFrame>=data.startFrame && data.endFrame>=endFrame) {
		return data.driver;
	    }
	}
	return null;
    }

    /**
     * Add a data manager listener.
     *
     * @param listener The listener
     */
    public void addListener(DataManagerListener listener)
    {
  	listeners.add(listener);
    }

    /**
     * Remove a listener.
     *
     * @param listener The listener
     */
    public void removeListener(DataManagerListener listener)
    {
  	listeners.remove(listener);
    }

    /**
     * Notify the listeners for a change in the data manager.
     *
     * @param e the event that describes the change
     */
    private void fireNotification(DataManagerEvent e)
    {
	for(Iterator i=listeners.iterator();i.hasNext();) {
	    DataManagerListener listener=(DataManagerListener)i.next();
	    listener.modelChanged(e);
	}
    }

    public void copyOffsetsFrom(DataManager otherManager)
    {
	offsets.clear();
	// Iterate and copy
	for(Iterator i=otherManager.offsets.keySet().iterator();i.hasNext();) {
	    Joint joint=(Joint)i.next();
	    // Get the matrix
	    Matrix4d result=otherManager.getOffset(joint);
	    offsets.put(joint,new Matrix4d(result));
	}
    }
    
    /**
     * Copy the drivers from another manager.
     *
     * @param otherManager The other manager.
     */
    public void copyDriversFrom(DataManager otherManager)
    {
	// Clean our lists
	rotationDrivers.clear();
	translationDrivers.clear();
	// Copy the data over
	copyDriverData(rotationDrivers,otherManager.rotationDrivers);
	copyDriverData(translationDrivers,otherManager.translationDrivers);
    }

    /**
     * Copy the driver data from one list to another.
     *
     * @param dest The destination list.
     * @param source The source list.
     */
    private void copyDriverData(List dest,List source)
    {
	for(Iterator i=source.iterator();i.hasNext();) {
	    DriverData driverData=(DriverData)i.next();
	    // Copy it over
	    DriverData newDriverData=new DriverData(driverData.driver,driverData.startFrame,driverData.endFrame);
	    //System.err.println(this+": Copying "+driverData.driver+","+driverData.startFrame+","+driverData.endFrame);
	    dest.add(newDriverData);
	}
    }
}

final class DriverData
{
    /** The driver */
    public final Object driver;
    
    /** The start frame for the driver */
    public final int startFrame;
    
    /** The end frame for the driver */
    public final int endFrame;
    
    /**
     * Initialize the driver info.
     *
     * @param driver The driver object.
     * @param startFrame The start frame for this driver.
     * @param endFrame The end frame for this driver.
     */
    public DriverData(Object driver,int startFrame,int endFrame)
    {
	this.driver=driver;
	this.startFrame=startFrame;
	this.endFrame=endFrame;
    }
}

