package mc.model;

import mc.util.Pair;
import com.sun.java.util.collections.*;
import javax.vecmath.Matrix4d;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * A class that alters the lengths of the limbs in order to 
 * retarget motion.
 */
public class GeometricRetargeter implements MotionRetargeter
{
    /**
     * A map of maps, indexed by root joint. The value Map
     * contains a mapping from ids to joints.
     */
    private final Map idMaps;
    
    public GeometricRetargeter()
    {
	this.idMaps=new HashMap();
    }
    
    /**
     * Retarget.
     *
     * @param model The animation model.
     * @param src The source skeleton.
     * @param target The target skeleton.
     */
    public void retarget(AnimationModel model,SkeletonHandle src,SkeletonHandle target)
    {
	SkeletonHandle source=src.getLODPyramid().getDefaultLevel();
	// Retrieve the id map of the source skeleton
	Map targetMap=new HashMap();
	buildMap(target.getRootJoint(),targetMap);
	// Clear the list
	Set mismatches=new HashSet();
	// Now, check which joints in the source
	// are missing from the target
	recordMismatches(mismatches,source.getRootJoint(),targetMap,true);
	// Now take care of the mismatches
	retargetMotion(source,target,model.getNumberOfFrames(),mismatches);
    }

    public void retarget(AnimationModel model,SkeletonHandle src,SkeletonHandle target,int frame)
    {
	throw new Error("NYI");
    }
    
    /**
     * Record any mismatches between the two skeletons.
     *
     * @param mismatches The value-result mismatches list.
     * @param joint The joint to examine.
     * @param targetMap The id map for the target skeleton.
     * @param parentExists If true, then the parent of this joint exists in the 
     * target skeleton.
     */
    private void recordMismatches(Set mismatches,Joint joint,Map targetMap,boolean parentExists)
    {
	Integer id=new Integer(joint.getId());
	// Check whether it exists in the target skeleton
	Joint targetJoint=(Joint)targetMap.get(id);
	if (targetJoint!=null && !parentExists) {
	    mismatches.add(targetJoint);
	}
	// Iterate over the children
	for(Iterator ci=joint.getChildren().iterator();ci.hasNext();) {
	    Joint childJoint=(Joint)ci.next();
	    recordMismatches(mismatches,childJoint,targetMap,(targetJoint!=null));
	}
    }


    /**
     * Retarget the motion of a set of joints, to match the
     * motion of the original joints.
     *
     * @param source The source skeleton.
     * @param target The target skeleton.
     * @param totalFrames The total number of frames.
     * @param mismatches The mismatches.
     */
    private void retargetMotion(SkeletonHandle source,SkeletonHandle target,int totalFrames,Set mismatches)
    {
	DataManager targetManager=target.getDataManager();
	TranslationDriver xlatDriver=targetManager.getTranslationDriverFor(0,totalFrames-1);
	TranslationDriver driver=new GeometricRetargetingTranslationDriver(xlatDriver,totalFrames,source,target,mismatches);
	target.getDataManager().addTranslationDriver(driver);
    }

    private void buildMap(Joint joint,Map map)
    {
	map.put(new Integer(joint.getId()),joint);
	// Recurse in the children
	for(Iterator ci=joint.getChildren().iterator();ci.hasNext();) {
	    Joint childJoint=(Joint)ci.next();
	    // Recurse
	    buildMap(childJoint,map);
	}
    }


}

		
