package mc.model;

import com.sun.java.util.collections.*;
import javax.vecmath.Matrix4d;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

public class GeometricRetargetingTranslationDriver implements TranslationDriver
{
    private final TranslationDriver driver;
    
    private final Map[] maps;

    private final SkeletonHandle source;
    
    private final SkeletonHandle target;
    
    private final Set retargetJoints;

    private final Map idMap;
    
    public GeometricRetargetingTranslationDriver(TranslationDriver driver,int numOfFrames,SkeletonHandle source,SkeletonHandle target,Set retargetJoints)
    {
	this.driver=driver;
	this.maps=new Map[numOfFrames];
	this.source=source;
	this.target=target;
	this.idMap=new HashMap();
	this.retargetJoints=retargetJoints;
	buildMapping(source.getRootJoint(),idMap);
    }
    
    public Matrix4d getTranslationMatrix(Joint joint,int frame)
    {
	// Get the original matrix
	Matrix4d originalMatrix=driver.getTranslationMatrix(joint,frame);

	// Check if we should retarget
	if (!retargetJoints.contains(joint)) {
	    return originalMatrix;
	}
	
	// First locate the map
	if (maps[frame]==null) {
	    maps[frame]=new HashMap();
	}
	// Try to find the matrix
	Matrix4d offset=(Matrix4d)maps[frame].get(joint);
	if (offset==null) {
	    offset=retarget(joint,frame);
	    maps[frame].put(joint,offset);
	}
	
	Matrix4d result=new Matrix4d(originalMatrix);
	result.mul(offset);
	return result;
    }
    
    private void buildMapping(Joint joint,Map map)
    {
	map.put(new Integer(joint.getId()),joint);
	// Recurse in the children
	for(Iterator ci=joint.getChildren().iterator();ci.hasNext();) {
	    Joint childJoint=(Joint)ci.next();
	    // Recurse
	    buildMapping(childJoint,idMap);
	}
    }

    /**
     * Get the orthonormal matrix for a joint, at a given frame number.
     *
     * @param joint The joint.
     * @param dataManager The data manager.
     * @param frame The frame.
     * @return The matrix
     */
    private Matrix4d getPosition(Joint joint,DataManager dataManager,int frame)
    {
	// Get the translation matrix for the node
	Matrix4d xlatMatrix=null;
	xlatMatrix=dataManager.getTranslationMatrix(joint,frame);
	
	// Get the rotation matrix
	Matrix4d rotMatrix=dataManager.getRotationMatrix(joint,frame);
	// 
	Matrix4d jointMatrix=new Matrix4d();
	jointMatrix.mul(xlatMatrix,rotMatrix);
	
	// Get the matrix of the parent
	Matrix4d parentMatrix=null;
	if (joint.getParent()!=null) {
	    // Get the parent's matrix
	    parentMatrix=getPosition(joint.getParent(),dataManager,frame);
	} else {
	    // Assign the identity matrix
	    parentMatrix=new Matrix4d();
	    parentMatrix.setIdentity();
	}
	// Calculate the final result
	Matrix4d result=new Matrix4d();
	result.mul(parentMatrix,jointMatrix);
	return result;
    }

    private Matrix4d retarget(Joint targetJoint,int frame) 
    {
	//System.err.println("Retargeting for "+targetJoint.getParent().getName()+"."+targetJoint.getName()+" at frame:"+frame);
	// Locate the source jont
	Joint sourceJoint=(Joint)idMap.get(new Integer(targetJoint.getId()));
	if (sourceJoint==null) {
	    throw new Error("Could not locate source for "+targetJoint.getName());
	}
	
	// Get the data managers
	DataManager sourceManager=source.getDataManager();
	DataManager targetManager=target.getDataManager();
	// Get the position of the parent of the first joint
	Matrix4d sourceMatrix=getPosition(sourceJoint.getParent(),sourceManager,frame);
	// Get the position of the second joint
	Matrix4d targetMatrix=getPosition(targetJoint.getParent(),targetManager,frame);
	// Get the translation of the source joint
	Matrix4d sourceXlat=new Matrix4d(sourceManager.getTranslationMatrix(sourceJoint,frame));
	// Get the translation matrix of the target joint
	Matrix4d targetXlat=new Matrix4d(driver.getTranslationMatrix(targetJoint,frame));
	targetXlat.mul(targetManager.getOffset(targetJoint));
	
	// Start the calculations
	targetMatrix.mul(targetXlat);
	sourceMatrix.mul(sourceXlat);
	// Transform two points
	Point3d sp=new Point3d(0,0,0);
	sourceMatrix.transform(sp);
	Point3d tp=new Point3d(0,0,0);
	targetMatrix.transform(tp);
	Point3d lp=new Point3d(sp);
	// Generate the displacement
	Matrix4d targetPrime=new Matrix4d(targetMatrix);
	targetPrime.invert();
	targetPrime.transform(sp);
	// Calculate the final matrix
	Matrix4d offset=new Matrix4d();
	offset.setIdentity();
	offset.set(new Vector3d(sp));
	// Test it
	Matrix4d test=new Matrix4d(targetMatrix);
	test.mul(offset);
	Point3d testpoint=new Point3d(0,0,0);
	test.transform(testpoint);
	return offset;
    }

    public DataRepresentation getTranslationRep(Joint joint,int frame)
    {
	Matrix4d xlat=getTranslationMatrix(joint,frame);
	TranslationVector3d v=new TranslationVector3d(0,0,0);
	xlat.get(v);
	return v;
    }
    
}




