package mc.model;

import com.sun.java.util.collections.*;
import java.awt.event.*;
import mc.view.RenderableView;

/**
 * Extend the BasicAnimationModel through an explicit pipeline setup.
 * The pipeline will determine what modifications are needed to the 
 * model to take LOD into account. It will be run for each model in the
 * scene and will take into account a master view whose camera and shape map
 * are used to determine its LOD. The distinct phases in the pipeline are to
 * choose the correct LOD for a model at the current time, do a geometric blending,
 * then retarget the motion.
 */
public class LODAnimationModel extends BasicAnimationModel
{
    /** The master view */
    private RenderableView   masterView = null;
    /** Component for geometric blending */
    private GeometryBlender  blender = null;
    /** Default Component for motion retargeting */
    private MotionRetargeter retargeter = null;
    /** The retargeters for each skeleton */
    private final Map retargeterMap;
    
    /**
     * @param timer
     * @param master view
     */
    public LODAnimationModel(ModelTimer timer, RenderableView masterView)
    {
	super(timer);
	this.masterView = masterView;
	this.blender = BlenderFactory.createDefaultBlender();
	//this.retargeter = RetargeterFactory.createDefaultRetargeter();
	this.retargeter = RetargeterFactory.createIKRetargeter();
	this.retargeterMap=new HashMap();
    }

    /**
     * Setup the pipeline before notifying listeners 
     * @param event
     */
    public void fireNotification(ModelEvent event)
    {
	setupLOD();
	super.fireNotification(event);
    }

    public void setRetargeter(SkeletonHandle handle,MotionRetargeter retargeter)
    {
	retargeterMap.put(handle,retargeter);
    }
    
    /**
     * Get a retargeter for a skeleton. If a designated retargeter has not
     * been assigned, then the default retargeter is returned.
     *
     * @param handle The skeleton handle.
     * @return The motion retargeter
     */
    private MotionRetargeter getRetargeter(SkeletonHandle handle)
    {
	MotionRetargeter designatedRetargeter=(MotionRetargeter)retargeterMap.get(handle);
	if (designatedRetargeter!=null) {
	    return designatedRetargeter;
	} else {
	    return retargeter;
	}
    }
    
    protected void notifyTick(int currentFrame,int increment)
    {
	setupLOD();
	super.notifyTick(currentFrame,increment);
    }
    
    /**
     * Setup the pipeline
     */
    protected void setupLOD()
    {
	// Check if we have a main view
	if (masterView==null) {
	    return;
	}
	
	Collection skeletons = getHandles();
	for(Iterator i = skeletons.iterator(); i.hasNext(); ) {
	    SkeletonHandle s = (SkeletonHandle)i.next();
	    SkeletonHandle sprime = s;
	    if(s.getLODPyramid() == null) {
		continue;
	    }
	    sprime = s.getLODPyramid().choose(masterView.getCamera(), getCurrentFrame());

	    // if new LOD...
	    if (sprime==null)
		continue;
	    
	    if(s.getRootJoint() != sprime.getRootJoint() ||
	       sprime.getRootJoint() != s.getLODPyramid().getDefaultLevel().getRootJoint()) {
		// Clone the skeleton so that the retargeter can mess with it.
		SkeletonHandle newSkeleton=createSkeleton(sprime.getRootJoint());
		newSkeleton.getDataManager().copyDriversFrom(sprime.getDataManager());
		newSkeleton.getDataManager().copyOffsetsFrom(sprime.getDataManager());
		// the retarget will be the first skeleton to the currently chosen skeleton
		blender.blend(newSkeleton, masterView, getCurrentFrame());

		if(sprime.getRootJoint() != s.getLODPyramid().getDefaultLevel().getRootJoint()) {
		    MotionRetargeter retargeter=getRetargeter(s);
		    retargeter.retarget(this,s, newSkeleton);
		}
		s.getDataManager().copyDriversFrom(newSkeleton.getDataManager());
		s.getDataManager().copyOffsetsFrom(newSkeleton.getDataManager());
		s.setRootJoint(newSkeleton.getRootJoint());
	    }
	}
    }
    
    /**
     * @param master view
     */
    public void setMasterView(RenderableView masterView)
    {
	this.masterView = masterView;
    }

    /**
     * @param blender
     */
    public void setBlender(GeometryBlender blender)
    {
	this.blender = blender;
    }

    /**
     * @param retargeter
     */
    public void setRetargeter(MotionRetargeter retargeter)
    {
	this.retargeter = retargeter;
    }

    /**
     * @return master view
     */
    public RenderableView getMasterView()
    {
	return masterView;
    }

    /**
     * @return blender
     */
    public GeometryBlender getBlender()
    {
	return blender;
    }

    /**
     * @return retargeter
     */
    public MotionRetargeter getRetargeter()
    {
	return retargeter;
    }
}
