package mc.model;

import javax.vecmath.Matrix4d;
import com.sun.java.util.collections.Map;

/**
 * A translation driver that stacks on top of another driver
 * and uses a mapping to get the matrices.
 */
public class MappedTranslationDriver implements TranslationDriver
{
    /** The real driver*/
    private final TranslationDriver theDriver;
    
    /** The map */
    private final Map theMap;
    
    /**
     * Initialize.
     *
     * @param theDriver The translation driver.
     * @param theMap The joint map.
     */
    public MappedTranslationDriver(TranslationDriver theDriver,Map theMap)
    {
	this.theDriver=theDriver;
	this.theMap=theMap;
    }
    
    /**
     * Get the translation matrix of the node.
     *
     * @param joint The joint.
     * @param frameNumber The frame number
     * @return The homogeneous 4x4 matrix
     */
    public Matrix4d getTranslationMatrix(Joint joint,int frameNumber)
    {
	// Look up the joint
	Joint mappedJoint=(Joint)theMap.get(joint);
	// Sanity check
	if (mappedJoint==null) {
	    throw new Error("Could not find mapping for joint "+joint.getName()+"," + joint);
	}
	return theDriver.getTranslationMatrix(mappedJoint,frameNumber);
    }

    /**
     * Get the internal representation of the translation.
     *
     * @return The internal representation.
     */
    public DataRepresentation getTranslationRep(Joint joint,int frameNumber)
    {
	// Look up the joint
	Joint mappedJoint=(Joint)theMap.get(joint);
	// Sanity check
	if (mappedJoint==null) {
	    throw new Error("Could not find mapping for joint "+joint);
	}
	return theDriver.getTranslationRep(mappedJoint,frameNumber);
    }
}


