package mc.model.euler;

import mc.model.MutableRotationDriver;
import mc.model.Joint;
import mc.model.FrameDataArray;
import mc.model.DataRepresentation;

import com.sun.java.util.collections.*;
import javax.vecmath.Vector3d;
import javax.vecmath.Matrix4d;

/**
 * The basic driver that uses precomputed data to drive
 * the whole process.
 */
public class BasicEulerRotationDriver implements EulerRotationDriver, MutableRotationDriver
{
    /** The map for rotation values */
    private final Map rotationValuesMap;
    
    /** The rotation matrix map */
    private Map rotationMatrixMap;

    /**
     * The map of String[3] for the rotation order of each node.
     */
    private final Map rotationOrderMap;
    
    /** The number of frames */
    private int numberOfFrames;
    
    static private final int ROTATEX=0;
    
    static private final int ROTATEY=1;
    
    static private final int ROTATEZ=2;
    
    /**
     * Initialize.
     *
     */
    public BasicEulerRotationDriver()
    {
	this.rotationValuesMap=new HashMap();
	this.rotationMatrixMap=new HashMap();
	this.rotationOrderMap=new HashMap();
    }

    public String getDescription()
    {
	return "Frame Driver";
    }
    
    /**
     * Initialize the driver for a number
     * frames.
     *
     * @param numberOfFrames The number of frames
     */
    public void initialize(int numberOfFrames)
    {
	rotationValuesMap.clear();
	this.numberOfFrames=numberOfFrames;
    }
    
    /**
     * Get the array for a joint. If the joint is not
     * associated with an array, one is created and registered.
     *
     * @param joint The joint.
     * @param map The map of joints->arrays
     * @return The array of the joint.
     */
    private FrameDataArray getArrayFor(Joint joint,Map map)
    {
	// Index in the map
	FrameDataArray array=(FrameDataArray)map.get(joint);
	if (array==null) {
	    array=new FrameDataArray(numberOfFrames);
	    map.put(joint,array);
	}
	return array;
    }

    /**
     * Worker method. Set the value in the array of a node, for
     * a specific frame number.
     *
     * @param joint The joint.
     * @param frameNumber The frame number
     * @param value The value.
     * @param arrayMap The map of arrrays
     * @param index The index of the value
     */
    private void setValue(Joint joint,int frameNumber,double value,Map arrayMap,int index)
    {
	// Look up the frame data for joint
	FrameDataArray frameData=getArrayFor(joint,arrayMap);
	// Set the appropriate value
	frameData.setValue(frameNumber,index,value);
    }

    /**
     * Worker method. Get the value of the array of 
     * a joint for the current frame.
     *
     * @param joint The joint.
     * @param frameNumber The frame number.
     * @param arrayMap The array map.
     * @param index The index.
     * @return the value.
     */
    private double getValue(Joint joint,int frameNumber,Map arrayMap,int index)
    {
	// Get the map
	FrameDataArray frameData=getArrayFor(joint,arrayMap);
	// Return the value
	return frameData.getValue(frameNumber,index);
    }

    /**
     * Set the rotateX for a joint, for a frame number.
     *
     * @param joint The joint.
     * @param frameNumber The frame number.
     * @param value The value.
     */
    public void setRotateX(Joint joint,int frameNumber,double value)
    {
	setValue(joint,frameNumber,value,rotationValuesMap,ROTATEX);
	// Update the rotation matrix
	Matrix4d[] matrices=getRotationMatrices(joint);
	//matrices[frameNumber].rotX(value);
	Matrix4d N = new Matrix4d();
	N.setIdentity();
	N.rotX(value);
	matrices[frameNumber].mul(N);
	// Add the rotation axis
	addRotationAxis(joint,X);
    }
    
    /**
     * Set the rotateY for a joint, for a frame number.
     *
     * @param joint The joint.
     * @param frameNumber The frame number.
     * @param value The value.
     */
    public void setRotateY(Joint joint,int frameNumber,double value)
    {
	setValue(joint,frameNumber,value,rotationValuesMap,ROTATEY);
	// Update the rotation matrix
	Matrix4d[] matrices=getRotationMatrices(joint);
	//matrices[frameNumber].rotY(value);
	Matrix4d N = new Matrix4d();
	N.setIdentity();
	N.rotY(value);
	matrices[frameNumber].mul(N);
	//
	addRotationAxis(joint,Y);
    }
    
    /**
     * Set the rotateZ for a joint, for a frame number.
     *
     * @param joint The joint.
     * @param frameNumber The frame number.
     * @param value The value.
     */
    public void setRotateZ(Joint joint,int frameNumber,double value)
    {
	setValue(joint,frameNumber,value,rotationValuesMap,ROTATEZ);
	// Update the rotation matrix
	Matrix4d[] matrices=getRotationMatrices(joint);
	//matrices[frameNumber].rotZ(value);
	Matrix4d N = new Matrix4d();
	N.setIdentity();
	N.rotZ(value);
	matrices[frameNumber].mul(N);
	//
	addRotationAxis(joint,Z);
    }
    
    /**
     * Get the rotateX channel of a node, for a given
     * frame number.
     *
     * @param joint The joint.
     * @param frameNumber The frame number.
     * @return The value of the channel.
     */
    public double getRotateX(Joint joint,int frameNumber)
    {
	return getValue(joint,frameNumber,rotationValuesMap,ROTATEX);
    }
    
   /**
     * Get the rotateY channel of a node, for a given
     * frame number.
     *
     * @param joint The joint.
     * @param frameNumber The frame number.
     * @return The value of the channel.
     */
    public double getRotateY(Joint joint,int frameNumber)
    {
	return getValue(joint,frameNumber,rotationValuesMap,ROTATEY);
    }

    /**
     * Get the rotateZ channel of a node, for a given
     * frame number.
     *
     * @param joint The joint.
     * @param frameNumber The frame number.
     * @return The value of the channel.
     */
    public double getRotateZ(Joint joint,int frameNumber)
    {
	return getValue(joint,frameNumber,rotationValuesMap,ROTATEZ);
    }

    /**
     * Get the rotation matrix of a joint for a specific frame.
     *
     * @param joint The joint.
     * @param frame The frame.
     * @return The 4x4 rotation matrix.
     */
    public Matrix4d getRotationMatrix(Joint joint,int frame)
    {
	Matrix4d[] matrices=getRotationMatrices(joint);
	return matrices[frame];
    }

    /**
     * Return the 3-Euler angle rotation vector.
     *
     * @param joint The joint.
     * @param frame The frame.
     * @return The EulerRotationVector3d
     */
    public DataRepresentation getRotationRep(Joint joint,int frame)
    {
	return new EulerRotationVector3d(getRotateX(joint,frame),getRotateY(joint,frame),getRotateZ(joint,frame),getRotationAxisOrder(joint));
    }

    /**
     * Get the rotation matrices for a specific joint. If the
     * matrices do not exist, a new array is created where all matrices
     * are set to the identity matrix.
     *
     * @param joint The joint.
     * @return The array of 4x4 matrices.
     */
    private Matrix4d[] getRotationMatrices(Joint joint)
    {
	return getMatrices(rotationMatrixMap,joint);
    }
    
    /**
     * Worker method. Check if the map contains the matrices for
     * a specific joint. If not, the array of matrices is created
     * and all matrices assigned the identity matrix.
     *
     * @param matricesMap The map to search.
     * @param joint The joint.
     * @return The array of 4x4 Matrices for the joint.
     */
    private Matrix4d[] getMatrices(Map matricesMap,Joint joint)
    {
	// Index in the map
	Matrix4d[] matrices=(Matrix4d[])matricesMap.get(joint);
	if (matrices==null) {
	    matrices=new Matrix4d[numberOfFrames];
	    for(int i=0;i<numberOfFrames;i++) {
		matrices[i]=new Matrix4d();
		matrices[i].setIdentity();
	    }

	    // Register it
	    matricesMap.put(joint,matrices);
	}
	return matrices;
    }
    
    /**
     * Get the order in which the euler angles have been applied.
     *
     * @param joint The joint.
     * @return An array of the symbols
     */
    public int[] getRotationAxisOrder(Joint joint)
    {
	return getOrderVector(joint);
    }

    /**
     * Add a rotation axis to the rotation axis order
     * of a joint.
     *
     * @param joint The joint.
     * @param axis The  acis
     * @exception IllegalArgumentException if the joint has already three axes in
     * the rotation order
     */
    private void addRotationAxis(Joint joint,int axis) throws IllegalArgumentException
    {
	// Get the vector
	int[] order=getOrderVector(joint);
	// Search for the first null
	int pos=0;
	for(;pos<order.length;pos++) {
	    if (order[pos]==-1) {
		order[pos]=axis;
		return;
	    }
	}
	return;
    }
    
    /**
     * Return the rotation vector associated with a 
     * joint. If no rotation vector exists, then one is 
     * created an initialized with nulls.
     *
     * @param joint The joint.
     * @return The rotation vector of the joint.
     */
    private int[] getOrderVector(Joint joint)
    {
	int[] rotVec=(int[])rotationOrderMap.get(joint);
	if (rotVec==null) {
	    rotVec=new int[3];
	    rotVec[0]=rotVec[1]=rotVec[2]=-1;
	    rotationOrderMap.put(joint,rotVec);
	}
	return rotVec;
    }
}
