package mc.model.euler;

import mc.model.RotationInterpolator;
import mc.model.RotationDriver;
import mc.model.Joint;
import mc.model.DataRepresentation;

import javax.vecmath.Matrix4d;

/**
 * A median filter implementation
 */
class EulerMedianFilterDriver implements RotationInterpolator, EulerRotationDriver
{
    /** The start frame */
    private int startFrame;
    
    /** The end frame */
    private int endFrame;
    
    /** The driver */
    private EulerRotationDriver theDriver;
    
    /**
     * Set up the interpolator
     *
     * @param startFrame The start frame
     * @param endFrame The end frame
     * @param driver The driver
     */
    public void interpolate(int startFrame,int endFrame,RotationDriver driver)
    {
	// Initialize the range of frames
	this.startFrame=startFrame;
	this.endFrame=endFrame;
	// Check that we will be using the correct driver
	if (!(driver instanceof EulerRotationDriver)) {
	    throw new IllegalArgumentException("Invalid driver "+driver);
	}
	this.theDriver=(EulerRotationDriver)driver;
    }
    
    /**
     * Get the rotation matrix for a joint, at a specified 
     * frame number.
     *
     * @param joint The joint.
     * @param frame The frame.
     * @return The matrix
     */
    public Matrix4d getRotationMatrix(Joint joint,int frame)
    {
	Matrix4d result=new Matrix4d();
	result.setIdentity();
	
	// Get the rotation order
	int[] rotationOrder=theDriver.getRotationAxisOrder(joint);
	for(int i=0;i<rotationOrder.length;i++) {
	    // Create a temp matrix
	    Matrix4d t=new Matrix4d();
	    // Check the axis
	    switch (rotationOrder[i]) {
	    case X:
		t.rotX(getRotateX(joint,frame));
		break;
	    case Y:
		t.rotY(getRotateY(joint,frame));
		break;
	    case Z:
		t.rotZ(getRotateZ(joint,frame));
		break;
	    default:
		// Must be an end-site
		if (joint.getChildren().size()!=0) {
		    throw new Error("Unknown axis "+rotationOrder[i]);
		}
	    }
	    result.mul(t);
	}
	return result;
    }

    public String getDescription()
    {
	return "Median Filter";
    }
    
    public int[] getRotationAxisOrder(Joint joint)
    {
	return theDriver.getRotationAxisOrder(joint);
    }
    
    /**
     * Calculate the median value
     *
     * @param v1 The t-1 value
     * @param v2 The t value
     * @param v3 The t+1 value
     * @return The median
     */
    private double calculateValue(double v1,double v2,double v3)
    {
	if (v1<v3) {
	  if (v1>v2) {
	    return v1;
	  }
	  else if (v2>v3) {
	    return v3;
	  } else {
	    return v2;
	  }
	}
	else if (v3>v2)  {
	  return v3;
	} else if (v2>v1) {
	  return v1;
	} else  {
	  return v2;
	}
    }

    /**
     * Return the 3-Euler angle rotation vector.
     *
     * @param joint The joint.
     * @param frame The frame.
     * @return The EulerRotationVector3d
     */
    public DataRepresentation getRotationRep(Joint joint,int frame)
    {
	return new EulerRotationVector3d(getRotateX(joint,frame),getRotateY(joint,frame),getRotateZ(joint,frame),getRotationAxisOrder(joint));
    }

    public double getRotateX(Joint joint,int frame)
    {
	if (frame==startFrame) 
	    return theDriver.getRotateX(joint,startFrame);
	else if (frame==endFrame)
	    return theDriver.getRotateX(joint,endFrame);
	else {
	    return calculateValue(theDriver.getRotateX(joint,frame-1),
				  theDriver.getRotateX(joint,frame),
				  theDriver.getRotateX(joint,frame+1));
	}
    }
    
    public double getRotateY(Joint joint,int frame)
    {
	if (frame==startFrame) 
	    return theDriver.getRotateY(joint,startFrame);
	else if (frame==endFrame)
	    return theDriver.getRotateY(joint,endFrame);
	else {
	    return calculateValue(theDriver.getRotateY(joint,frame-1),
				  theDriver.getRotateY(joint,frame),
				  theDriver.getRotateY(joint,frame+1));
	}

    }

    public double getRotateZ(Joint joint,int frame)
    {
	if (frame==startFrame) 
	    return theDriver.getRotateZ(joint,startFrame);
	else if (frame==endFrame)
	    return theDriver.getRotateZ(joint,endFrame);
	else {
	    return calculateValue(theDriver.getRotateZ(joint,frame-1),
				  theDriver.getRotateZ(joint,frame),
				  theDriver.getRotateZ(joint,frame+1));
	}
    }
    
}

		

