package mc.model.euler;

import mc.model.RotationInterpolator;
import mc.model.RotationDriver;
import mc.model.Joint;
import mc.model.DataRepresentation;

import javax.vecmath.Matrix4d;

/**
 * The interpolator that uses Euler angles
 */
class EulerRotationInterpolator implements RotationInterpolator, EulerRotationDriver
{
    /** The start frame */
    private int startFrame;
    
    /** The end frame */
    private int endFrame;
    
    /** The driver */
    private EulerRotationDriver theDriver;
    
    /**
     * Set up the interpolator
     *
     * @param startFrame The start frame
     * @param endFrame The end frame
     * @param driver The driver
     */
    public void interpolate(int startFrame,int endFrame,RotationDriver driver)
    {
	// Initialize the range of frames
	this.startFrame=startFrame;
	this.endFrame=endFrame;
	// Check that we will be using the correct driver
	if (!(driver instanceof EulerRotationDriver)) {
	    throw new IllegalArgumentException("Invalid driver "+driver);
	}
	this.theDriver=(EulerRotationDriver)driver;
    }
    
    /**
     * Get the rotation matrix for a joint, at a specified 
     * frame number.
     *
     * @param joint The joint.
     * @param frame The frame.
     * @return The matrix
     */
    public Matrix4d getRotationMatrix(Joint joint,int frame)
    {
	Matrix4d result=new Matrix4d();
	result.setIdentity();
	
	// Get the rotation order
	int[] rotationOrder=theDriver.getRotationAxisOrder(joint);
	for(int i=0;i<rotationOrder.length;i++) {
	    // Create a temp matrix
	    Matrix4d t=new Matrix4d();
	    // Check the axis
	    switch (rotationOrder[i]) {
	    case X:
		t.rotX(getRotateX(joint,frame));
		break;
	    case Y:
		t.rotY(getRotateY(joint,frame));
		break;
	    case Z:
		t.rotZ(getRotateZ(joint,frame));
		break;
	    default:
		// Must be an end-site
		if (joint.getChildren().size()!=0) {
		    throw new Error("Unknown axis "+rotationOrder[i]);
		}
	    }
	    result.mul(t);
	}
	return result;
    }

    public String getDescription()
    {
	return "Linear Interpolation";
    }
    
    public int[] getRotationAxisOrder(Joint joint)
    {
	return theDriver.getRotationAxisOrder(joint);
    }
    
    /**
     * Calculate the linear interpolation between two values.
     *
     * @param start the start value
     * @param end the end value
     * @param frame the frame
     * @return the linearly interpolated value
     */
    private double calculateValue(double start,double end,int frame)
    {
	return start+(end-start)*(frame-startFrame)/((double)endFrame-startFrame);
    }

    /**
     * Return the 3-Euler angle rotation vector.
     *
     * @param joint The joint.
     * @param frame The frame.
     * @return The EulerRotationVector3d
     */
    public DataRepresentation getRotationRep(Joint joint,int frame)
    {
	return new EulerRotationVector3d(getRotateX(joint,frame),getRotateY(joint,frame),getRotateZ(joint,frame),getRotationAxisOrder(joint));
    }

    public double getRotateX(Joint joint,int frame)
    {
	return calculateValue(theDriver.getRotateX(joint,startFrame),
			      theDriver.getRotateX(joint,endFrame),
			      frame);
    }
    
    public double getRotateY(Joint joint,int frame)
    {
	return calculateValue(theDriver.getRotateY(joint,startFrame),
			      theDriver.getRotateY(joint,endFrame),
			      frame);
    }

    public double getRotateZ(Joint joint,int frame)
    {
	return calculateValue(theDriver.getRotateZ(joint,startFrame),
			      theDriver.getRotateZ(joint,endFrame),
			      frame);
    }
    
}

		
