package mc.model.map;

import mc.model.quart.BasicQuartenionRotationDriver;
import mc.model.MutableRotationDriver;
import mc.model.Joint;
import mc.model.DataRepresentation;

import javax.vecmath.Quat4d;
import javax.vecmath.Matrix4d;

/**
 * A driver that uses exponential maps. The actual implementation uses
 * a core driver for storing the rotations in the form of quartenions
 * and converting back to the exponential map.
 */
public class BasicMapRotationDriver implements MutableRotationDriver, MapRotationDriver
{
    /**
     * The core driver
     */
    private final BasicQuartenionRotationDriver coreDriver;
    
    /**
     * Initialize.
     *
     */
    public BasicMapRotationDriver()
    {
	this.coreDriver=new BasicQuartenionRotationDriver();
    }

    public String getDescription()
    {
	return "Frame Driver";
    }
    
    /**
     * Initialize for the number of frames.
     *
     * @param numberOfFrames The number of frames
     */
    public void initialize(int numberOfFrames)
    {
	coreDriver.initialize(numberOfFrames);
    }
    
    /**
     * Set the X rotation for a joint and a specific frame.
     *
     * @param joint The joint
     * @param frameNumber The frame number
     * @param x the rotation
     */
    public void setRotateX(Joint joint,int frameNumber,double x)
    {
	coreDriver.setRotateX(joint,frameNumber,x);
    }
    
    /**
     * Set the Y rotation for a joint and a specific frame.
     *
     * @param joint The joint
     * @param frameNumber The frame number
     * @param y the rotation
     */
    public void setRotateY(Joint joint,int frameNumber,double y)
    {
	coreDriver.setRotateY(joint,frameNumber,y);
    }

    /**
     * Set the Z rotation for a joint and a specific frame.
     *
     * @param joint The joint
     * @param frameNumber The frame number
     * @param z the rotation
     */
    public void setRotateZ(Joint joint,int frameNumber,double z)
    {
	coreDriver.setRotateZ(joint,frameNumber,z);
    }

    public ParameterizedQuat4d getParameterizedQuat4d(Joint joint,int frameNumber)
    {
	// Get the quartenion of the node
	Quat4d quat=coreDriver.getQuartenionFor(joint,frameNumber);
	// Convert it throught the exponential map
	ParameterizedQuat4d result=new ParameterizedQuat4d();
	result.set(quat);
	return result;
    }
    

    /**
     * Get the rotaion matrix for a joint and a specific frame.
     * This entails getting the parameterized quartenion for the joint,
     * converting it to a unit-length quartenion, and calculating the matrix
     *
     * @param joint The joint
     * @param frameNumber The frame number
     * @return The rotation matrix
     */
    public Matrix4d getRotationMatrix(Joint joint,int frameNumber)
    {
	// Get the map
	ParameterizedQuat4d pQuat=getParameterizedQuat4d(joint,frameNumber);
	// Get the matrix
	return pQuat.getMatrix4d();
    }

    /**
     * Get the data representation for a parameterized quartenion.
     *
     * @param joint The joint.
     * @param frameNumber The frame.
     * @return The parameterized quartenion
     */
    public DataRepresentation getRotationRep(Joint joint,int frameNumber)
    {
	return getParameterizedQuat4d(joint,frameNumber);
    }
}

