package mc.model.map;

import javax.vecmath.Vector3d;
import javax.vecmath.Quat4d;
import javax.vecmath.AxisAngle4d;
import javax.vecmath.Matrix4d;

import mc.model.DataRepresentation;

/**
 * A parameterized quartenion
 */
public class ParameterizedQuat4d extends Vector3d implements DataRepresentation
{
    /** The name for this encoding */
    public final static String Name="Parameterized Quartenion";
    
    /**
     * Initialize the parameterized quartenion to 
     * a rotation represented as an axis angle rotation
     *
     * @param rotation The rotation
     */
    public ParameterizedQuat4d(Quat4d rotation)
    {
	super();
	// Convert to a quartenion
	set(rotation);
    }
    
    /**
     * Initialize to the null quat
     *
     */
    public ParameterizedQuat4d()
    {
	super();
    }
    
    /**
     * Set the map to the rotation described by a quartenion
     *
     * @param quat The quartenion that describes the rotation
     */
    public void set(Quat4d quat)
    {
	if (quat.x==0 && quat.y==0 && quat.z==0) {
	    x=0;
	    y=0;
	    z=0;
	    return;
	}
	
	// Extract the axis
	Vector3d axis=new Vector3d(quat.x,quat.y,quat.z);
	// Calculate the arc cosine
	double acos=2*Math.acos(quat.w);
	// Scale the axis
	axis.scale(acos/axis.length());
	// Set our representaton
	super.set(axis);
    }

    /**
     * Return the equivalent normalized quartenion
     * from this map.
     *
     * @return The quartenion
     */
    public Quat4d getQuat4d()
    {
	if (x==0 && y==0 && z==0)
	    return new Quat4d(0,0,0,1);
	
	// Calculate the theta
	double theta=length();
	// Calculate the sinc
	double sinc=sincTerm(theta);
	// Calculate the axis of the quartenion
	Vector3d axis=new Vector3d(x,y,z);
	// Normalize the axis
	axis.scale(sinc);
	// Return the quartenion
	Quat4d result=new Quat4d(axis.x,axis.y,axis.z,Math.cos(theta/2));
	return result;
    }

    /**
     * Calculate the sinc(theta/2) function through the Taylor expansion.
     * This will use the first three terms only.
     *
     * @param t The theta
     * @return The sinc(t/2)/t
     */
    private double sincTerm(double t)
    {
	double theta=t/2;
	
	double thetaSqr=theta*theta;
	double thetaQu=thetaSqr*thetaSqr;
	return 0.5+thetaSqr/48-thetaQu/3840;
    }
    
    /**
     * Get the rotation matrix that corresponds to this
     * parameterized quartenion
     *
     * @return The matrix
     */
    public Matrix4d getMatrix4d()
    {
	// Get the quartenion
	Quat4d quat=getQuat4d();
	// Get the matrix out of that
	Matrix4d result=new Matrix4d();
	result.set(quat);
	return result;
    }
}


	
