package mc.model.quart;

import mc.model.MutableRotationDriver;
import mc.model.Joint;
import mc.model.DataRepresentation;

import com.sun.java.util.collections.*;
import javax.vecmath.*;

/**
 * the basic quartenion driver
 */
public class BasicQuartenionRotationDriver implements MutableRotationDriver, QuartenionRotationDriver
{
    /**
     * The map with the quartenions
     */
    private final Map quatMap;

    /** the number of frames */
    private int numberOfFrames;
    
    /**
     * Initialize
     *
     */
    public BasicQuartenionRotationDriver()
    {
	this.quatMap=new HashMap();
    }
    
    /**
     * Record the # of frames 
     *
     * @param numberOfFrames the number of frames.
     */
    public void initialize(int numberOfFrames)
    {
	this.numberOfFrames=numberOfFrames;
    }

    public String getDescription()
    {
	return "Frame Driver";
    }
    
    /**
     * Set the rotateX channel. This will multiply the quartenion of
     * the node to a normalized quartenion that represents the rotation 
     * around the x-axis.
     *
     * @param joint The joint.
     * @param frameNumber The frame number
     * @param x The rotation
     */
    public void setRotateX(Joint joint,int frameNumber,double x)
    {
	// Get the quartenion for the node
	Quat4d[] quats=getQuartenions(joint);
	
	// Form a new quartenion
	AxisAngle4d rot=new AxisAngle4d(1,0,0,x);
	Quat4d xQuat=new Quat4d();
	xQuat.set(rot);
	// Multiply the new quartenion with the existing quat
	quats[frameNumber].mul(xQuat);
    }

    /**
     * Set the rotateY channel. This will multiply the quartenion of
     * the node to a normalized quartenion that represents the rotation 
     * around the y-axis.
     *
     * @param joint The joint.
     * @param frameNumber The frame number
     * @param y The rotation
     */
    public void setRotateY(Joint joint,int frameNumber,double y)
    {
	// Get the quartenion for the node
	Quat4d[] quats=getQuartenions(joint);
	
	// Form a new quartenion
	AxisAngle4d rot=new AxisAngle4d(0,1,0,y);
	Quat4d yQuat=new Quat4d();
	yQuat.set(rot);
	// Multiply the new quartenion with the existing quat
	quats[frameNumber].mul(yQuat);
    }
    
    /**
     * Set the rotateZ channel. This will multiply the quartenion of
     * the node to a normalized quartenion that represents the rotation around the z-axis.
     *
     * @param joint The joint.
     * @param frameNumber The frame number
     * @param z The rotation
     */
    public void setRotateZ(Joint joint,int frameNumber,double z)
    {
	// Get the quartenion for the node
	Quat4d[] quats=getQuartenions(joint);
	
	// Form a new quartenion
	AxisAngle4d rot=new AxisAngle4d(0,0,1,z);
	Quat4d zQuat=new Quat4d();
	zQuat.set(rot);
	// Multiply the new quartenion with the existing quat
	quats[frameNumber].mul(zQuat);
    }

    /**
     * Worker method. Get the array of quartenions for one node.
     *
     * @param joint the joint
     * @return the array of quartenions for the node.
     */
    private Quat4d[] getQuartenions(Joint joint)
    {
	Quat4d[] quats=(Quat4d[])quatMap.get(joint);
	if (quats==null) {
	    quats=new Quat4d[numberOfFrames];
	    for(int i=0;i<numberOfFrames;i++) {
		quats[i]=new Quat4d(0,0,0,1);
	    }
	    quatMap.put(joint,quats);
	}
	return quats;
    }

    /**
     * Return the matrix that corresponds to the quartenion
     * at a specific frame.
     *
     * @param joint the joint
     * @param frameNumber the frame number
     * @return the rotation matix
     */
    public Matrix4d getRotationMatrix(Joint joint,int frameNumber)
    {
	Quat4d[] quats=getQuartenions(joint);
	Matrix4d result=new Matrix4d();
	result.set(quats[frameNumber]);
	return result;
    }
    
    /**
     * Get the quartenion for a joint and a frame.
     *
     * @param joint The joint
     * @param frameNumber The frame number
     * @return The quartenion
     */
    public Quat4d getQuartenionFor(Joint joint,int frameNumber)
    {
	Quat4d[] quats=getQuartenions(joint);
	return quats[frameNumber];
    }
    
    public DataRepresentation getRotationRep(Joint joint,int frameNumber)
    {
	return new QuartenionRepresentation(getQuartenionFor(joint,frameNumber));
    }
}
