package mc.util;

import javax.swing.*;
import com.sun.java.util.collections.*;
import java.awt.event.*;

/**
 * A button that sends acceleration events when pushed.
 */
public class AccelButton extends JButton implements MouseListener, ActionListener
{
    /**
     * Default constructor.
     *
     */
    public AccelButton()
    {
	super();
	init();
    }
    
    /**
     * Construct a button with an icon.
     *
     * @param icon The icon.
     */
    public AccelButton(Icon icon)
    {
	super(icon);
	init();
    }
    
    /**
     * Construct a button with a label.
     *
     * @param text The text of the label.
     */
    public AccelButton(String text)
    {
	super(text);
	init();
    }
    
    /**
     * Construct a button with a label and an icon.
     *
     * @param text The label.
     * @param icon The icon.
     */
    public AccelButton(String text,Icon icon)
    {
	super(text,icon);
	init();
    }
    
    /** The default delay between successive firings of the accel event */
    private final static int DEFAULT_DELAY=100;
    
    /** The timer used for the acceleration events */
    private Timer timer;
    
    /**
     * Initialize the timer with the default delay
     *
     */
    private void init()
    {
	this.timer=new Timer(100,this);
	this.listeners=new LinkedList();
	super.addMouseListener(this);
    }
    
    /**
     * Set the delay between successive accel events.
     *
     * @param delay the delay.
     */
    public void setDelay(int delay)
    {
	this.timer.setDelay(delay);
    }
    
    /**
     * The list of listeners
     */
    private List listeners;
    
    /**
     * Add an acceleration listener. 
     *
     * @param listener The acceleration listener.
     */
    public void addAccelListener(AccelListener listener)
    {
	listeners.add(listener);
    }
    
    /**
     * Remove an acceleration listener. 
     *
     * @param listener The listener.
     */
    public void removeAccelListener(AccelListener listener)
    {
	listeners.remove(listener);
    }
    
    /**
     * Notify the listeners of an acceleration event.
     *
     * @param e The event.
     */
    protected void fireAccelerationEvent(ActionEvent e)
    {
	// Iterate over the listeners
	for(Iterator i=listeners.iterator();i.hasNext();) {
	    // Get the listener
	    AccelListener listener=(AccelListener)i.next();
	    // Notify
	    listener.accelerated(e);
	}
    }

    
    /**
     * Notify the listeners of an acceleration event.
     *
     * @param e The event.
     */
    protected void fireStopEvent(ActionEvent e)
    {
	// Iterate over the listeners
	for(Iterator i=listeners.iterator();i.hasNext();) {
	    // Get the listener
	    AccelListener listener=(AccelListener)i.next();
	    // Notify
	    listener.stopped(e);
	}
    }
    
    /**
     * The mouse was pressed. Start the timer.
     *
     * @param e The event
     */
    public void mousePressed(MouseEvent e) 
    {
	// Start the timer
	timer.start();
    }
    
    /**
     * Stop the timer.
     *
     * @param e The mouse event.
     */
    public void mouseReleased(MouseEvent e)
    {
	// Stop the timer
	timer.stop();
	// Fire the stop event
	fireStopEvent(new ActionEvent(this,0,getActionCommand()));
    }
    
    public void mouseClicked(MouseEvent e)
    {}
    
    public void mouseEntered(MouseEvent e)
    {}
    
    public void mouseExited(MouseEvent e)
    {}
    
    /**
     * The timer has expired. Check if the button is armed
     * and fire an acceleration event.
     *
     * @param e The action event.
     */
    public void actionPerformed(ActionEvent e)
    {
	// Check if the button is armed
	if (getModel().isArmed()) {
	    fireAccelerationEvent(new ActionEvent(this,0,getActionCommand()));
	}
    }

    public static void  main(String[] args)
    {
	JFrame frame=new JFrame();
	AccelButton b=new AccelButton("Press me");
	b.setDelay(100);
	frame.getContentPane().add(b);
	b.addAccelListener(new TestListener());
	frame.setSize(100,100);
	frame.pack();
	frame.setVisible(true);
	frame.addWindowListener(new WindowAdapter(){
		public void windowClosing(WindowEvent e) 
		{
		    System.exit(0);
		}
	    });
    }
}

class TestListener implements AccelListener
{
    public void stopped(ActionEvent e)
    {
	System.err.println("Event stopped");
    }
    
    public void accelerated(ActionEvent e)
    {
	System.err.println("Accelerated");
    }
}


