package mc.util;

import mc.model.*;
import mc.view.*;
import javax.vecmath.*;
import com.sun.java.util.collections.*;

/**
 * Collection of useful scene utilities
 */

public final class SceneUtil
{
    public static final int XMIN = 0;
    public static final int XMAX = 1;
    public static final int YMIN = 2;
    public static final int YMAX = 3;
    public static final int ZMIN = 4;
    public static final int ZMAX = 5;
    
    /**
     * Return the orientation of a joint given to be the 
     * rotation of k^.
     * @return orientation vector
     */
    public static final Vector3d getJointOrientation(SkeletonHandle skeleton, Joint joint, int frame)
    {
	Matrix4d m = findGlobalJointMatrix(skeleton, new HashMap(), joint, frame);
	// drop the translation component
	m.setTranslation(new Vector3d());
	// assume the default orientation vector is -k', rotate it
	Vector3d orientation = new Vector3d(0,0,1);
	m.transform(orientation);
	return orientation;
    }

    /**
     * Return the world space matrix of a given joint for a given Skeleton
     * @param SkeletonHandle
     * @param matrix map
     * @param desiredjoint
     * @param desired frame
     * @return matrix
     */
    public static final Matrix4d findGlobalJointMatrix(SkeletonHandle skeleton, Map mCache,
						       Joint joint, int frame)
    {
	Matrix4d m = (Matrix4d)mCache.get(joint);
	if(m != null) {
	    return m;
	}
	m = new Matrix4d();
	m.setIdentity();	
	Matrix4d t = skeleton.getDataManager().getTranslationMatrix(joint, frame);
	Matrix4d r = skeleton.getDataManager().getRotationMatrix(joint, frame);
	m.mul(t,r);

	if(joint.getParent() == null) {
	    mCache.put(joint, m);
	    return m;
	}
	Matrix4d ancestors = new Matrix4d(findGlobalJointMatrix(skeleton, mCache, joint.getParent(), frame));
	ancestors.mul(m);
	mCache.put(joint, ancestors);
	return ancestors;
    }

    /**
     * Find the bounding volume of the scene
     * @param AnimationMode
     * @param desired time
     * @return Pair of bbox and offset
     */
    public static final Pair getBoundingVolume(AnimationModel model, int frame)
    {
	double[] max = null;
	for(Iterator i = model.getHandles().iterator(); i.hasNext(); ) {
	    SkeletonHandle skeleton = (SkeletonHandle)i.next();
	    double[] current = new double[] {Double.MAX_VALUE, Double.MIN_VALUE,
					     Double.MAX_VALUE, Double.MIN_VALUE,
					     Double.MAX_VALUE, Double.MIN_VALUE};
	    findBounds(skeleton, new HashMap(), skeleton.getRootJoint(), frame, current);
	    if(max == null) {
		max = current;
	    } else {
		// expand current box if need be
		if(current[XMIN] < max[XMIN])
		    max[XMIN] = current[XMIN];
		if(current[XMAX] > max[XMAX])
		    max[XMAX] = current[XMAX];
		if(current[YMIN] < max[YMIN])
		    max[YMIN] = current[YMIN];
		if(current[YMAX] > max[YMAX])
		    max[YMAX] = current[YMAX];
		if(current[ZMIN] < max[ZMIN])
		    max[ZMIN] = current[ZMIN];
		if(current[ZMAX] > max[ZMAX])
		    max[ZMAX] = current[ZMAX];
	    }
	}
	return new Pair(new BoundingVolume(max[XMAX] - max[XMIN],
					   max[ZMAX] - max[ZMIN],
					   max[YMAX] - max[YMIN]),
			new Vector3d(max[XMIN] + (max[XMAX]-max[XMIN])/2,
				     max[YMIN] + (max[YMAX]-max[YMIN])/2,
				     max[ZMIN] + (max[ZMAX]-max[ZMIN])/2));
    }

    /**
     * Clone a skeleton. This entails creating a new skeleton with the same joint
     * hierarchy, cloned drivers and a cloned Pyramid.
     *
     * @param model The model that will contain the clone.
     * @param handle The handle to clone.
     * @return The new handle.
     */
    public static SkeletonHandle cloneSkeletonSameJoints(BasicAnimationModel model,SkeletonHandle handle)
    {
	SkeletonHandle clone=model.addSkeleton(handle.getRootJoint());
	// Copy over the drivers
	clone.getDataManager().copyOffsetsFrom(handle.getDataManager());
	clone.getDataManager().copyDriversFrom(handle.getDataManager());
	// Now clone the pyramid
	LODPyramid pyramid=new LODPyramid(null);
	clone.setLODPyramid(pyramid);
	for(Iterator i=handle.getLODPyramid().getLevels();i.hasNext();) {
	    SkeletonHandle lodHandle=(SkeletonHandle)i.next();
	    SkeletonHandle lodClone=new SkeletonHandle(lodHandle.getRootJoint(),new DataManager());
	    lodClone.getDataManager().copyOffsetsFrom(lodHandle.getDataManager());
	    lodClone.getDataManager().copyDriversFrom(lodHandle.getDataManager());
	    if (lodHandle==handle.getLODPyramid().getDefaultLevel()) {
		clone.getLODPyramid().addDefaultLevel(lodClone);
	    } else {
		clone.getLODPyramid().addLevel(lodClone);
	    }
	}
	// Return the clone
	return clone;
    }

    /**
     * Clone a skeleton. This entails creating a new skeleton with the same joint
     * hierarchy, cloned drivers and a cloned Pyramid.
     *
     * @param model The model that will contain the clone.
     * @param handle The handle to clone.
     * @return The new handle.
     */
    public static SkeletonHandle cloneSkeletonSameJoints(BasicAnimationModel model,SkeletonHandle handle,Map sourceMap,Map targetMap)
    {
	SkeletonHandle clone=model.addSkeleton(handle.getRootJoint());
	// Copy over the drivers
	clone.getDataManager().copyOffsetsFrom(handle.getDataManager());
	clone.getDataManager().copyDriversFrom(handle.getDataManager());
	// Now clone the pyramid
	LODPyramid pyramid=new LODPyramid(null);
	clone.setLODPyramid(pyramid);
	for(Iterator i=handle.getLODPyramid().getLevels();i.hasNext();) {
	    SkeletonHandle lodHandle=(SkeletonHandle)i.next();
	    SkeletonHandle lodClone=new SkeletonHandle(lodHandle.getRootJoint(),new DataManager());
	    lodClone.getDataManager().copyOffsetsFrom(lodHandle.getDataManager());
	    lodClone.getDataManager().copyDriversFrom(lodHandle.getDataManager());
	    targetMap.put(lodClone,sourceMap.get(lodHandle));
	    if (lodHandle==handle.getLODPyramid().getDefaultLevel()) {
		clone.getLODPyramid().addDefaultLevel(lodClone);
	    } else {
		clone.getLODPyramid().addLevel(lodClone);
	    }
	}
	// Return the clone
	return clone;
    }
    
    /**
     * Clone a skeleton in a model and add a translation to the hip joint.
     * 
     * @param model The model that will contain the cloned skeleton.
     * @param handle The skeleton that will be cloned.
     * @param translation The additional translation of the root joint.
     * @return The cloned skeleton.
     */
    public static SkeletonHandle cloneSkeletonSameJoints(BasicAnimationModel model,SkeletonHandle handle,Vector3d translation)
    {
	// Clone it
	SkeletonHandle clone=cloneSkeletonSameJoints(model,handle);
	// Now get every driver and add a translation to the root joint
	addDisplacement(0,model.getNumberOfFrames(),clone,translation);
	for(Iterator i=clone.getLODPyramid().getLevels();i.hasNext();) {
	    SkeletonHandle lodHandle=(SkeletonHandle)i.next();
	    addDisplacement(0,model.getNumberOfFrames(),lodHandle,translation);
	}
	return clone;
    }
    
    /**
     * Add a displacement to the root joint of a skeleton.
     *
     * @param startFrame The starting frame.
     * @param endFrame The ending frame.
     * @param handle The skeleton.
     * @param translation The displacement.
     */
    public static void addDisplacement(int startFrame,int endFrame,SkeletonHandle handle,Vector3d translation)
    {
	// Get the data manger
	DataManager dataManager=handle.getDataManager();
	// Get the driver for the frames - Assumes that we will stack on top of the 
	// first driver
	TranslationDriver xlatDriver=dataManager.getTranslationDriverFor(startFrame,endFrame);
	dataManager.addTranslationDriver(startFrame,endFrame,new TransformTranslationDriver(translation,xlatDriver));
    }
    
    /**
     * Find the bounding volume of a skeleton. Return the bbox and global
     * displacement in a Pair of BoundingVolume and Vector3d types.
     * @param SkeletonHandle
     * @param desired frame
     * @return bbox and displacement
     */
    public static final Pair getBoundingVolume(SkeletonHandle skeleton, int frame)
    {
	// min,max for x,y,z
	double[] bounds = new double[] {Double.MAX_VALUE, Double.MIN_VALUE,
					     Double.MAX_VALUE, Double.MIN_VALUE,
					     Double.MAX_VALUE, Double.MIN_VALUE};

	findBounds(skeleton, new HashMap(), skeleton.getRootJoint(), frame, bounds);

	return new Pair(new BoundingVolume(bounds[XMAX] - bounds[XMIN],
					   bounds[ZMAX] - bounds[ZMIN],
					   bounds[YMAX] - bounds[YMIN]),
			new Vector3d(bounds[XMIN] + (bounds[XMAX] - bounds[XMIN])/2,
				     bounds[YMIN] + (bounds[YMAX] - bounds[YMIN])/2,
				     bounds[ZMIN] + (bounds[ZMAX] - bounds[ZMIN])/2));
    }

    /**
     * Helper method for finding the bounding volume of skeleton. 
     * Updates the passed in bound relative to the given joint.
     * @param skeleton
     * @param matrix map
     * @param desired joint
     * @param desired frame
     * @param current and next bound
     */
    public static final void findBounds(SkeletonHandle skeleton, HashMap mCache,
					Joint joint, int frame, double[] bounds)
    {
	Point3d origin = new Point3d();
	findGlobalJointMatrix(skeleton, mCache, joint, frame).transform(origin);
	if(origin.x < bounds[XMIN]) {
	    bounds[XMIN] = origin.x;
	} 
	if(origin.x > bounds[XMAX]) {
	    bounds[XMAX] = origin.x;
	}
	if(origin.y < bounds[YMIN]) {
	    bounds[YMIN] = origin.y;
	} 
	if(origin.y > bounds[YMAX]) {
	    bounds[YMAX] = origin.y;
	}
	if(origin.z < bounds[ZMIN]) {
	    bounds[ZMIN] = origin.z;
	} 
	if(origin.z > bounds[ZMAX]) {
	    bounds[ZMAX] = origin.z;
	}
	for(Iterator i = joint.getChildren().iterator(); i.hasNext();) {
	    Joint child = (Joint)i.next();
	    findBounds(skeleton, mCache, child, frame, bounds);
	}
    }

    public static double[] glIdentityMatrix = new double[] {    
	1.0, 0.0, 0.0, 0.0,
	0.0, 1.0, 0.0, 0.0,
	0.0, 0.0, 1.0, 0.0,
	0.0, 0.0, 0.0, 1.0};

    public static double[] toGLMatrix(Matrix4d m)
    {
	double[] glm = glIdentityMatrix;
	if(m != null) {
	    glm = new double[16];
	    glm[0] = m.m00;glm[1] = m.m10;glm[2] = m.m20;glm[3] = m.m30;
	    glm[4] = m.m01;glm[5] = m.m11;glm[6] = m.m21;glm[7] = m.m31;
	    glm[8] = m.m02;glm[9] = m.m12;glm[10] = m.m22;glm[11] = m.m32;
	    glm[12] = m.m03;glm[13] = m.m13;glm[14] = m.m23;glm[15] = m.m33;	
	}
	return glm;
    }   

}
