package mc.util;

import mc.view.Camera;
import mc.model.ModelAdapter;
import mc.model.AnimationModel;
import mc.model.TimerListener;
import mc.model.ModelTimer;
import mc.model.TimerEvent;
import mc.view.RenderableView;
import javax.vecmath.Vector3d;

/**
 * A swinging camera director
 */
public class SwingingCameraController extends ModelAdapter
{
    /** The camera to control */
    private RenderableView view;
    
    /** The starting point */
    private final Vector3d startingPoint;
    
    /** The ending point */
    private final Vector3d endingPoint;
    
    /** The number of steps for the cycle */
    private final int cycleSteps;
    
    /** The current step */
    private int currentStep;

    /** Records whether the controller is returning to the starting point */
    private boolean returning;
    
    /**
     * Initialize the director.
     *
     * @param model The model that this controller is attached to. 
     * @param startingPoint The starting point of the camera.
     * @param endingPoint The ending point of the camera.
     * @param cycle The number of frames it takes to perform half a trip.
     */
    public SwingingCameraController(AnimationModel model,Vector3d startingPoint, Vector3d endingPoint,int cycle)
    {
	model.addModelListener(this);
	this.startingPoint=startingPoint;
	this.endingPoint=endingPoint;
	this.cycleSteps=cycle;
	this.currentStep=cycleSteps;
	this.returning=false;
    }
    
    /**
     * Set the camera to control.
     *
     * @param camera The camera to control.
     */
    public void setView(RenderableView view)
    {
	this.view=view;
    }
    
    public void frameTicked(int currentFrame,int increment)
    {
	Vector3d start=null;
	Vector3d end=null;
	double alpha;
	
	// Check if a new cycle should start
	if (currentStep==cycleSteps) {
	    returning=!returning;
	    currentStep=0;
	}
	
	if (!returning) {
	    start=new Vector3d(startingPoint);
	    end=new Vector3d(endingPoint);
	} else {
	    start=new Vector3d(endingPoint);
	    end=new Vector3d(startingPoint);
	} 
	
	// Calculate the alpha
	alpha=currentStep/(double)(cycleSteps-1);
	
	end.sub(start);
	end.scale(alpha);
	start.add(end);
	
	// Set the coordinates now
	Camera camera=view.getCamera();
	camera.setEyeX(start.x);
	camera.setEyeY(start.y);
	camera.setEyeZ(start.z);

	currentStep++;
    }
}

