package mc.view;

import mc.model.AnimationModel;
import mc.model.Joint;
import mc.model.SkeletonHandle;
import mc.model.LODPyramid;
import mc.model.euler.*;
import mc.view.cameras.*;
import mc.view.shapes.*;
import mc.util.Pair;
import mc.util.SceneUtil;

import javax.swing.*;
import java.awt.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import javax.vecmath.Matrix4d;
import javax.vecmath.Vector3d;
import javax.vecmath.Quat4d;
import javax.vecmath.Color4f;
import com.sun.java.util.collections.*;
import com.hermetica.magician.*;
import com.hermetica.util3d.*;

/**
 * The default tree view.
 */
public class GLModelView extends JPanel implements RenderableView, GLEventListener
{
    /** The animation model */
    private AnimationModel model;

    private GL gl = null;
    private GLU glu = null;
    private CoreGL core_gl;
    private CoreGLU core_glu;
    private TraceGL trace_gl;
    private GLComponent glc = null;
    private ShapeManager smgr;
    private ShapeManager tsmgr;
    private Vector shapeManagers;
    private int numTraceFrames;
    private String cameraType;
    private Camera camera;
    private AutoCamera debugCamera;
    private JPanel internal;
    private String prefix = "default";
    private boolean isRecord = false;
    private boolean cameraDebugFlag = true;
    private int fileCount=0;
    
    private SceneProperties sceneProperties;

    /**
     * Initialize
     *
     */
    public GLModelView(String cameraType)
    {
	// Vertical split
	super();
	core_gl = new CoreGL();
	core_glu = new CoreGLU();
	trace_gl = new TraceGL(core_gl);
	gl = core_gl;
	glu = core_glu;
	glc = (GLComponent)GLComponentFactory.createGLComponent(200, 200);
	smgr = new ShapeManager();
	tsmgr = new ShapeManager();
	shapeManagers = new Vector();
	shapeManagers.addElement(smgr);
	shapeManagers.addElement(tsmgr);
	numTraceFrames = 0;
	this.cameraType = cameraType;
	sceneProperties = new SceneProperties();
    }
    
    public void setModel(AnimationModel model)
    {
	this.model=model;
	if(cameraType == null) {
	    camera = new ManualCamera();
	    debugCamera = new AutoCamera(DirectorFactory.createLocalDirector(DirectorFactory.CHOPPY), model);	    
	} else {
	    camera = new AutoCamera(DirectorFactory.createLocalDirector(DirectorFactory.CHOPPY), model);
	}

	// Add the gl component
	setLayout( new BorderLayout());
	add(BorderLayout.CENTER, glc);
	internal = new JPanel(new BorderLayout());
	internal.add(BorderLayout.NORTH, new GLRecordComponent(this));
	internal.add(BorderLayout.CENTER, new GLCameraComponent(this));
	internal.add(BorderLayout.SOUTH, new GLTraceDepthComponent(this));
	loadShapes();
    }
    
    private void loadShapes()
    {
	// initialize shapes, including those in pyramid
	for(Iterator i = model.getHandles().iterator(); i.hasNext();) {	    
	    SkeletonHandle s = (SkeletonHandle)i.next();
	    initShapes(s, s.getRootJoint());
	    LODPyramid p = s.getLODPyramid();
	    if(p!=null) {
		for(Iterator j = p.getLevels(); j.hasNext();) {
		    SkeletonHandle ss = (SkeletonHandle)j.next();
		    if(ss != s) {
			initShapes(ss, ss.getRootJoint());
		    }
		}
	    }
	}
    }

    public void initView()
    {
	// setup the capabilities and init
	GLCapabilities cap = glc.getContext().getCapabilities();
	cap.setDoubleBuffered(GLCapabilities.DOUBLEBUFFER);
	cap.setDepthBits(12);
	cap.setColourBits(24);
	cap.setPixelType(GLCapabilities.RGBA);
	glc.addGLEventListener(this);
	glc.initialize();  


    }

    public Component getComponent()
    {
	return this;
    }

    public Component getControlComponent()
    {
	return internal;
    }
    
    public Camera getCamera()
    {
	return camera;
    }

    public void setCameraDebug(boolean flag)
    {
	cameraDebugFlag = flag;
    }

    public Iterator getShapeManagers()
    {
	return shapeManagers.iterator();	
    }

    public AnimationModel getModel()
    {
	return model;
    }

    public SceneProperties getSceneProperties()
    {
	return sceneProperties;
    }

    /** Just call repaint */
    public void refresh()
    {
	// poke gl to redraw redraw iteself
	glc.repaint();
    }
    
    public void display(GLDrawable component) 
    {
	gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
	gl.glEnable(GL.GL_DEPTH_TEST);
	gl.glColor3f(1f, 1f, 1f);
	gl.loadIdentity();
	
	// look at scene
	Vector3d eye = camera.getEye(model.getCurrentFrame());
	Vector3d tgt = camera.getTarget(model.getCurrentFrame());
	glu.gluLookAt(eye.x,eye.y,eye.z,tgt.x,tgt.y,tgt.z,0,1,0);

	// draw defualt scene objects
	sceneProperties.drawIndependentObjects(gl);

	// draw tracers
	int start = 0;
	if(numTraceFrames >= 0)
	    start = Math.max(0,model.getCurrentFrame()-numTraceFrames);
	for(int i = model.getCurrentFrame(); i > start; --i) {
	    traverseAllHierarchies(tsmgr, i);
	}
	// draw skeletons
	traverseAllHierarchies(smgr, model.getCurrentFrame());
	
	// debug camera
	if(cameraDebugFlag) {
	    if(debugCamera != null)
		debugCamera.draw(model, gl);
	    // camera
	    camera.draw(model, gl);
	}

	if(isRecord) {
	    PPMWriter writer = new PPMWriter((GLComponent)component, "lalakis");
      	    String[] commands={
      		"sh","-c","ppmtotga lalakis"+fileCount+".ppm > "+prefix+fileCount+".tga ; rm lalakis"+fileCount+".ppm"
      	    };

	    try {
		Process p=Runtime.getRuntime().exec(commands);
		System.err.println(p.waitFor());
	    }
	    catch(Throwable t){
		System.err.println(t);
	    }
	    fileCount++;
	}
    }
    
    public GL getGL() 
    {
	return gl;
    }

    public void initialize(GLDrawable d) 
    {
	gl.glClearColor(0f, 0f, 0f, 1.0f);
	sceneProperties.setupLights(gl);
	sceneProperties.setupMaterials(gl);
    }

    public void reloadScene()
    {
	gl.glClearColor(0f, 0f, 0f, 1.0f);
	sceneProperties.setupLights(gl);
	sceneProperties.setupMaterials(gl);
	loadShapes();
    }
    
    public void reshape(GLDrawable component, int x, int y, int width, int height) 
    {
	gl.glViewport(component, x,y,width, height);
	gl.glMatrixMode(GL.GL_PROJECTION);
	gl.glLoadIdentity();
	//gl.glOrtho(-100.0,100.0,-100.0,100.0,-100.0,100.0);
	glu.gluPerspective(45,1, 100.0,1000.0);
	//	gl.glTranslatef(0,0,10.0f);
	gl.glMatrixMode(GL.GL_MODELVIEW);
    }

    public void setNumTraceFrames(int numTraceFrames)
    {
	this.numTraceFrames = numTraceFrames;
	refresh();
    }

    public int getNumTraceFrames()
    {
	return numTraceFrames;
    }

    public void setRecordPrefix(String prefix)
    {
	System.out.println(prefix);
	this.prefix = prefix;
    }

    public String getRecordPrefix()
    {	
	return prefix;
    }

    public void setRecord(boolean isRecord)
    {
	this.isRecord = isRecord;
    }

    private void traverseAllHierarchies(ShapeManager mgr, int frame)
    {
	Collection c = model.getHandles();
	for(Iterator i = c.iterator(); i.hasNext();) {
	    SkeletonHandle s = (SkeletonHandle)i.next();
	    traverseHierarchy(s.getRootJoint(), s, mgr, frame);
	}
    }

    private void traverseHierarchy(Joint joint, SkeletonHandle handle, 
				   ShapeManager mgr, int frame)
    {
	// set up current stack
	Matrix4d r = handle.getDataManager().getRotationMatrix(joint, frame);
	Matrix4d x = handle.getDataManager().getTranslationMatrix(joint, frame);
	gl.glPushMatrix();

	Vector3d xdata = new Vector3d();
	x.get(xdata);       

	gl.glTranslated(xdata.x, xdata.y, xdata.z);
	gl.glMultMatrixd(SceneUtil.toGLMatrix(r));

	// draw it
	JointAdapter ja = new JointAdapter(joint);
	for(Iterator i = mgr.getJointShapes(ja); i.hasNext();) {
	    ((Shape)i.next()).draw(ja, gl);
	}
	
	// descend
	for(Iterator i = joint.getChildren().iterator(); i.hasNext(); ) {
	    Joint child = (Joint)i.next();
	    // do segment from parent to child
	    SegmentAdapter sa = new SegmentAdapter(joint, child, handle, 3.0, frame);	    
	    gl.glPushMatrix();	    
	    gl.glMultMatrixd(SceneUtil.toGLMatrix(sa.getLocalMatrix()));
	    for(Iterator j = mgr.getSegmentShapes(sa); j.hasNext();) {
		Shape s = (Shape)j.next();
		s.draw(sa,gl);
	    }
	    gl.glPopMatrix();

	    traverseHierarchy(child, handle, mgr, frame);
	}
	// cleanup
	gl.glPopMatrix();
    }

    private void initShapes(SkeletonHandle skeleton, Joint joint)
    {
	smgr.addShapeToJoint(joint, ShapeManager.DEFAULT_JOINT_SHAPE);
	// descend
	for(Iterator i = joint.getChildren().iterator(); i.hasNext(); ) {
	    Joint child = (Joint)i.next();
	    if(sceneProperties.getSkeletonSegmentShape(skeleton) != null) {
		smgr.addShapeToSegment(joint, child,sceneProperties.getSkeletonSegmentShape(skeleton));
	    } else{
		smgr.addShapeToSegment(joint, child, ShapeManager.DEFAULT_SEGMENT_SHAPE);
	    }
	    initShapes(skeleton, child);
	}
	// add tracers for end effectors
	if(joint.getChildren().size() == 0) {
	    tsmgr.addShapeToJoint(joint, ShapeManager.DEFAULT_JOINT_SHAPE);
	}
    }
}

