package mc.view;

import com.hermetica.magician.*;
import com.hermetica.util3d.*;
import mc.model.DataRepresentation;
import mc.model.quart.QuartenionRepresentation;
import mc.util.FormatBigDecimal;
import javax.vecmath.Quat4d;
import javax.swing.*;
import java.awt.*;

/**
 * A class that shows a quartenion in a GL view
 */
public class QuartenionGLView extends JPanel implements DataRepresentationView, GLEventListener
{
    /** The quat to display */
    private Quat4d quat;
    
    /** The GL drawing context */
    private final GL gl;

    /** The glu context */
    private final GLU glu;
    
    /** The component */
    private final GLComponent component;

    /** The angle field */
    private JTextField angleField;
    
    /** The display list */
    private int displayList;
    
    /**
     * Initialize 
     *
     */
    public QuartenionGLView()
    {
	super(new BorderLayout());
	
	displayList=-1;
	
	// Get a new engine
	gl=new CoreGL();
	glu=new CoreGLU();
	
	// Create the component
	component = (GLComponent)GLComponentFactory.createGLComponent(100, 100);
	// Initialize the text field
	JPanel anglePanel=new JPanel();
	anglePanel.add(new JLabel("Angle: "));
	angleField=new JTextField(7);
	angleField.setEditable(false);
	angleField.setHorizontalAlignment(JTextField.RIGHT);
	anglePanel.add(angleField);
	// Add the component
	super.add(BorderLayout.CENTER,component);
	super.add(BorderLayout.SOUTH,anglePanel);
    }

    public GL getGL()
    {
	return gl;
    }
    
    public void reshape(GLDrawable drawable,int x,int y,int width,int height)
    {
	System.err.println("Reshaped");
	gl.glViewport(drawable, x,y,width, height);
	checkError("After viewport");
	gl.glMatrixMode(GL.GL_PROJECTION);
	checkError("After Projection matrix mode");
	gl.glLoadIdentity();
	checkError("After Load Identity");
	gl.glOrtho(-100.0,100.0,-100.0,100.0,-100.0,100.0);
	checkError("After Ortho");
	//glu.gluPerspective(45,1, 100.0,400.0);
	// gl.glTranslatef(0,0,10.0f);
	gl.glMatrixMode(GL.GL_MODELVIEW);
	checkError("After Model mode");
    }

    private void checkError(String msg)
    {
	switch(gl.glGetError()) {
	case GL.GL_NO_ERROR:
	    System.err.println(msg+": no error");
	    break;
	case GL.GL_INVALID_ENUM:
	    System.err.println(msg+": invalid enum");
	    break;
	case GL.GL_INVALID_VALUE:
	    System.err.println(msg+": invalid value");
	    break;
	case GL.GL_INVALID_OPERATION:
	    System.err.println(msg+": invalid id operation");
	    break;
	case GL.GL_STACK_OVERFLOW:
	    System.err.println(msg+": stack overflow");
	    break;
	case GL.GL_STACK_UNDERFLOW:
	    System.err.println(msg+": stack underflow");
	    break;
	case GL.GL_OUT_OF_MEMORY:
	    System.err.println(msg+": out of memory");
	    break;
	}
    }
    
    public void initView()
    {
	// setup the capabilities and init
	GLCapabilities cap = component.getContext().getCapabilities();
	cap.setDoubleBuffered(GLCapabilities.DOUBLEBUFFER);
	cap.setDepthBits(12);
	cap.setColourBits(24);
	cap.setPixelType(GLCapabilities.RGBA);
	component.addGLEventListener(this);
	component.initialize();
    }
    
    public void repaint()
    {
	super.repaint();
    }
    
    public void initialize(GLDrawable component) 
    {
	gl.glClearColor(0f, 0f, 0f, 1.0f);
	// Build the lists
	displayList = gl.glGenLists( 1 );
	// Build the circle
	GLUQuadric qobj=glu.gluNewQuadric();
	glu.gluQuadricDrawStyle(qobj,GLU.GLU_LINE);
        glu.gluQuadricNormals( qobj, GLU.GLU_NONE );
	// Start the list
        gl.glNewList( displayList, GL.GL_COMPILE );
	glu.gluDisk(qobj,79,80,20,1);
	gl.glEndList();
	// End the list
    }

    /**
     * Refresh the gl view with the new information.
     *
     * @param rep The representation of a quartenion
     */
    public void refresh(DataRepresentation rep)
    {
	// Downcast to a quartenion
	this.quat=(QuartenionRepresentation)rep;
	// Repaint
	component.repaint();
    }
    
    public void display(GLDrawable drawable)
    {
	gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
	gl.glEnable(GL.GL_DEPTH_TEST);
	gl.glColor3f(1f, 1f, 1f);
	gl.loadIdentity();

	gl.glRotatef(10,0f,1f,0f);
	gl.glRotatef(10,1f,0f,0f);
	
	double min_x = 0;
	double max_x = 80;
	double min_y = -3;
	double max_y = 3;
	double min_z = -3;
	double max_z = 3;

	// Draw the two axis
	gl.glPushMatrix();
	gl.glColor3f(1.0f,1.0f,1.0f);
	gl.glBegin(GL.GL_LINES);
	gl.glVertex2d(-80,0);
	gl.glVertex2d(80,0);
	gl.glVertex3d(0,0,-80);
	gl.glVertex3d(0,0,80);
	gl.glEnd();
	gl.glPopMatrix();
	
	// Draw the disk
	gl.glPushMatrix();
	gl.glColor3d(0.0,1.0,0.0);
	gl.glCallList(displayList);
	gl.glPopMatrix();
	
	gl.glColor3d(1.0,0.0,0.0);
	gl.glPushMatrix();
	gl.glRotatef(90f,0f,1f,0f);
	gl.glCallList(displayList);
	gl.glPopMatrix();

	gl.glColor3f(0.0f,1.0f,1.0f);
	gl.glPushMatrix();
	gl.glRotatef(90,1,0,0);
	gl.glCallList(displayList);
	gl.glPopMatrix();

	if (quat==null) {
	    angleField.setText("Error");
	    return;
	}

	double angle=(180/Math.PI)*2*Math.acos(quat.w);
	angleField.setText(FormatBigDecimal.format(angle,2).toString());
	
	// Calculate the azimuth and rotation from the quartenion
	double theta=Math.atan(quat.z/quat.x);
	if (Double.isNaN(theta)) {
	    theta=90;
	} else {
	    theta*=(180/Math.PI);
	}
	
	double phi=Math.atan(quat.y/Math.sqrt(Math.pow(quat.x,2)+Math.pow(quat.z,2)));
	if (Double.isNaN(phi)) {
	    phi=90;
	} else {
	    phi*=(180/Math.PI);
	}


	gl.glPushMatrix();
	gl.glRotated(theta,0,1,0);
	gl.glRotated(phi,0,0,1);
	gl.glColor3d(1.0, 0.0, 0.0);	
	gl.glBegin(GL.GL_QUADS);
	  gl.glVertex4d(min_x, max_y, max_z, 1.0);
	  gl.glVertex4d(min_x, min_y, max_z, 1.0);
	  gl.glVertex4d(max_x, min_y, max_z, 1.0);
	  gl.glVertex4d(max_x, max_y, max_z, 1.0);
  	gl.glEnd();
  	gl.glBegin(GL.GL_QUADS);
	  gl.glVertex4d(min_x, max_y, min_z, 1.0);
	  gl.glVertex4d(min_x, min_y, min_z, 1.0);
	  gl.glVertex4d(max_x, min_y, min_z, 1.0);
	  gl.glVertex4d(max_x, max_y, min_z, 1.0);
	gl.glEnd();
	gl.glColor3d(0.0, 1.0, 0.0);	
	gl.glBegin(GL.GL_QUADS);
	  gl.glVertex4d(min_x, max_y, min_z, 1.0);
	  gl.glVertex4d(min_x, max_y, max_z, 1.0);
	  gl.glVertex4d(max_x, max_y, max_z, 1.0);
	  gl.glVertex4d(max_x, max_y, min_z, 1.0);
	gl.glEnd();
	gl.glBegin(GL.GL_QUADS);
	  gl.glVertex4d(min_x, min_y, min_z, 1.0);
	  gl.glVertex4d(min_x, min_y, max_z, 1.0);
	  gl.glVertex4d(max_x, min_y, max_z, 1.0);
	  gl.glVertex4d(max_x, min_y, min_z, 1.0);
	gl.glEnd();
	gl.glColor3d(0.0, 0.0, 1.0);	
	gl.glBegin(GL.GL_QUADS);
	  gl.glVertex4d(min_x, max_y, max_z, 1.0);
	  gl.glVertex4d(min_x, max_y, min_z, 1.0);
	  gl.glVertex4d(min_x, min_y, min_z, 1.0);
	  gl.glVertex4d(min_x, min_y, max_z, 1.0);
	gl.glEnd();
	gl.glBegin(GL.GL_QUADS);
	  gl.glVertex4d(max_x, max_y, max_z, 1.0);
	  gl.glVertex4d(max_x, max_y, min_z, 1.0);
	  gl.glVertex4d(max_x, min_y, min_z, 1.0);
	  gl.glVertex4d(max_x, min_y, max_z, 1.0);
	gl.glEnd();
	gl.glPopMatrix();

    }

    public Component getComponent()
    {
	return this;
    }
    
}

    
