package mc.view;

import mc.model.*;
import mc.view.shapes.*;
import javax.vecmath.*;
import com.sun.java.util.collections.*;
import com.hermetica.magician.*;

/**
 * Manages odds and ends for the properties of scene.
 * These include:
 * 1. Lighting
 * 2. Shapes independent of AM's (sa floor)
 * 3. AM part to shape setting
 */
public class SceneProperties
{
    private HashMap skeletonSegmentShapeMap;
    private HashMap skeletonFatnessMap;
    private HashMap skeletonJointShapeMap;
    private HashMap skeletonTraceShapeMap;

    private Vector independentObjects;

    private LightInfo[] lights;
    private MaterialInfo materialInfo;

    public SceneProperties()
    {
	skeletonSegmentShapeMap = new HashMap();
	skeletonFatnessMap  = new HashMap();
	skeletonJointShapeMap = new HashMap();
	skeletonTraceShapeMap = new HashMap();

	independentObjects = new Vector();

	lights = new LightInfo[8]; // max number of lights in GL is 8
	materialInfo = null;
    }

    public void setSkeletonSegmentToShape(SkeletonHandle skeleton, Shape shape)
    {
	skeletonSegmentShapeMap.put(skeleton, shape);
    }

    public void setSkeletonFatness(SkeletonHandle skeleton, double fatness)
    {
	skeletonFatnessMap.put(skeleton, new Double(fatness));
    }

    public void setSkeletonJointToShape(SkeletonHandle skeleton, Shape shape)
    {
	skeletonJointShapeMap.put(skeleton, shape);
    }

    public void setSkeletonTraceToShape(SkeletonHandle skeleton, Shape shape)
    {
	skeletonTraceShapeMap.put(skeleton, shape);
    }

    public Shape getSkeletonSegmentShape(SkeletonHandle skeleton)
    {
	return (Shape)skeletonSegmentShapeMap.get(skeleton);
    }

    public Double getSkeletonFatness(SkeletonHandle skeleton)
    {
	return (Double)skeletonFatnessMap.get(skeleton);
    }

    public Shape getSkeletonJointShape(SkeletonHandle skeleton)
    {
	return (Shape)skeletonJointShapeMap.get(skeleton);
    }

    public Shape getSkeletonTraceShape(SkeletonHandle skeleton)
    {
	return (Shape)skeletonTraceShapeMap.get(skeleton);
    }

    public void addIndependentObject(ShapeAdapter adapter, Shape shape, Color4f color, Matrix4d matrix)
    {
	independentObjects.addElement(new IndependentShape(adapter, shape, color, matrix));
    }

    public Iterator getIndependentObjects()
    {
	return independentObjects.iterator();
    }

    public void drawIndependentObjects(GL gl)
    {
	for(Iterator i = getIndependentObjects(); i.hasNext(); ) {
	    ((IndependentShape)i.next()).draw(null, gl);
	}
    }

    public void addLightInfo(Vector3f position,
			     Vector3f ambientColor,
			     Vector3f specularColor,
			     int index)
    {
	if(index < lights.length && index >= 0) {
	    lights[index] = new LightInfo(position, ambientColor, specularColor);
	}
    }

    public void setMaterialInfo(Vector3f specularReflectionColor, 
				int shininess)
    {
	materialInfo = new MaterialInfo(specularReflectionColor, shininess);
    }

    public void setupLights(GL gl)
    {
	for(int i = 0; i < lights.length; i++) {
	    if(lights[i] != null)
		lights[i].setup(gl, i);
	}
    }

    public void setupMaterials(GL gl)
    {
	if(materialInfo != null)
	    materialInfo.setup(gl);
    }

    class LightInfo
    {
	private float[] al;
	private float[] sp;
	private float[] pos;

	public LightInfo(Vector3f position, 
			 Vector3f ambientColor,
			 Vector3f specularColor)
	{
	    pos = new float[]{position.x, position.y, position.z, 1.0f};
	    al  = new float[]{ambientColor.x, ambientColor.y, ambientColor.z};
	    sp  = new float[]{specularColor.x, specularColor.y, specularColor.z};
	}

	public void setup(GL gl, int index)
	{
	    int lindex = GL.GL_LIGHT0;
	    switch(index) {
	    case 0:		
		lindex = GL.GL_LIGHT0;
		break;
	    case 1:
		lindex = GL.GL_LIGHT1;
		break;
	    case 2:
		lindex = GL.GL_LIGHT2;
		break;
	    case 3:
		lindex = GL.GL_LIGHT3;
		break;
	    case 4:
		lindex = GL.GL_LIGHT4;
		break;
	    case 5:
		lindex = GL.GL_LIGHT5;	
		break;    
	    case 6:
		lindex = GL.GL_LIGHT6;
		break;
	    case 7:
		lindex = GL.GL_LIGHT7;
		break;
	    default:
		break;
	    }
	    gl.glEnable(GL.GL_LIGHTING);
	    gl.glLightModelfv(GL.GL_LIGHT_MODEL_AMBIENT, al);
	    gl.glLightfv(lindex, GL.GL_DIFFUSE, al);
	    gl.glLightfv(lindex, GL.GL_SPECULAR, sp);
	    gl.glLightfv(lindex, GL.GL_POSITION, pos);
	    gl.glEnable(lindex);
	}
    }

    class MaterialInfo
    {
	private float[] spr;
	private int shininess;

	public MaterialInfo(Vector3f specularReflectionColor, 
			    int shininess)
	{
	    spr = new float[]{specularReflectionColor.x,
				  specularReflectionColor.y,
				  specularReflectionColor.z};
	    this.shininess = shininess;
	}

	public void setup(GL gl)
	{
	    gl.glEnable(GL.GL_COLOR_MATERIAL);
	    gl.glColorMaterial(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE);
	    gl.glMaterialfv(GL.GL_FRONT, GL.GL_SPECULAR, spr);
	    gl.glMateriali(GL.GL_FRONT, GL.GL_SHININESS, shininess);
	}
    }
}
