package mc.view;

import javax.vecmath.Matrix4d;
import javax.vecmath.Vector3d;
import javax.vecmath.Tuple3d;
import javax.vecmath.AxisAngle4d;
import javax.vecmath.Point3d;

import mc.model.AnimationModel;
import mc.model.SkeletonHandle;
import mc.model.Joint;
import mc.util.Pair;

/**
 * Produce data generated by a Segment. A segment is
 * inferred from its parent and child joints and contains
 * its own local transform relative to the parent.
 */
public class SegmentAdapter implements ShapeAdapter
{
    private Joint parent;
    private Joint child;
    private BoundingVolume volume;
    private Matrix4d matrix;

    public SegmentAdapter(Joint parent, Joint child, SkeletonHandle skeleton, double fatness, int frame)
    {
	this.parent = parent;
	this.child = child;
	this.volume = new BoundingVolume(1.0, 1.0*fatness, 1.0*fatness);
	calcBoundingVolume(skeleton, frame);
	calcMatrix(skeleton, frame);
    }
    public SegmentAdapter(Joint parent, Joint child, SkeletonHandle skeleton, int frame)
    {
	this(parent, child, skeleton, 1.0, frame);
    }

    public BoundingVolume getBoundingVolume()
    {
	return volume;
    }

    public Matrix4d getLocalMatrix()
    {
	return matrix;
    }

    public Joint getParent()
    {
	return parent;
    }

    public Joint getChild()
    {
	return child;
    }
    
    // let the length of the bounding volume be the distance between
    // parent and child joints
    private void calcBoundingVolume(SkeletonHandle skeleton, int frame)
    {
	Matrix4d tc = skeleton.getDataManager().getTranslationMatrix(child, frame);
	
	double dx = tc.m03;
	double dy = tc.m13;
	double dz = tc.m23;

	volume.setLength(Math.sqrt(dx*dx + dy*dy + dz*dz));
    }

    // let the local segment transform be the offset to the midpoint
    // between parent and child joints
    private void calcMatrix(SkeletonHandle skeleton, int frame)
    {
	Matrix4d tc = skeleton.getDataManager().getTranslationMatrix(child, frame);

	matrix = new Matrix4d();
	matrix.setIdentity();
	
	Matrix4d xlat=new Matrix4d();
	xlat.setIdentity();
	xlat.m03=volume.getLength()/2;
	xlat.m13=0;
	xlat.m23=0;
	
//  	matrix.m03 = tc.m03/2.0;
//  	matrix.m13 = tc.m13/2.0;
//  	matrix.m23 = tc.m23/2.0;	    

	Pair angles = null;
	angles = calcPolar(new Vector3d(tc.m03,
					    tc.m13,
					    tc.m23));

	// System.err.println(angles);
	
	// The theta rotation
	Matrix4d rz = new Matrix4d();
	rz.setIdentity();
	rz.rotZ(((Double)angles.first).doubleValue());
	
	// Figure out the phi rotation
	Matrix4d ry = new Matrix4d();
	ry.setIdentity();
//  	Point3d yUnit=new Point3d(0,1,0);
//  	rz.transform(yUnit);
//  	AxisAngle4d phiAxis=new AxisAngle4d(yUnit.x,yUnit.y,yUnit.z,-((Double)angles.second).doubleValue());
//  	ry.set(phiAxis);
	ry.rotY(-((Double)angles.second).doubleValue());

	matrix.mul(rz);
	matrix.mul(ry);
	matrix.mul(xlat);
    }

    private Pair calcPolar(Tuple3d tail)
    {
	Pair angles = new Pair();
	
	double theta = 0;
	if(tail.x > 0 && tail.y > 0)
	    theta = Math.atan(tail.y/tail.x);
	else if(tail.x < 0)
	    theta = Math.PI + Math.atan(tail.y/tail.x);
	else if(tail.x > 0 && tail.y < 0)
	    theta = 2*Math.PI + Math.atan(tail.y/tail.x);
	else if(tail.x == 0 && tail.y >= 0)
	    theta = Math.PI/2.0;
	else if(tail.x == 0 && tail.y < 0)
	    theta = 3*Math.PI/2.0;
	else if (tail.y==0) {
	    theta=0;
	} else {
	    throw new Error("Not covered");
	}
	
	double phi = 0;
	double xyproj=tail.x*tail.x+tail.y*tail.y;
	if(xyproj != 0) {
	    phi = Math.atan(tail.z/(Math.sqrt(xyproj)));
	    if (phi<0) {
		phi+=2*Math.PI;
	    }
	}
	else if(tail.x == 0 && tail.y == 0 && tail.z > 0)
	    phi = Math.PI/2.0;
	else if(tail.x == 0 && tail.y == 0 && tail.z < 0)
	    phi =3*Math.PI/2.0;
	else if (tail.z==0) {
	    phi=0;
	} else {
	    throw new Error("Not covered");
	}
	
	angles.first = new Double(theta);
	angles.second = new Double(phi);
	return angles;
    }
}
