package mc.view;

import mc.model.AnimationModel;
import mc.model.Joint;
import mc.model.SkeletonHandle;
import mc.util.Pair;

import javax.swing.*;
import java.awt.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import com.sun.java.util.collections.*;

/**
 * The default tree view.
 */
public class TreeModelView extends JPanel implements ModelView , TreeSelectionListener
{
    /** The animation model */
    private AnimationModel model;
    
    /**
     * The tree.
     */
    private JTree theTree;

    /** The view for rotation data */
    private DataRepresentationView rotView;
    
    /** The joint that is currently selected */
    private Joint currentJoint;
    
    /** The current handle */
    private SkeletonHandle currentHandle;
    
    /**
     * Initialize
     *
     */
    public TreeModelView()
    {
	// Vertical split
	super(new BorderLayout());
	theTree=null;
    }
    
    public void setModel(AnimationModel model)
    {
	this.model=model;
	initTreeNodes();

	// Add the tree
	JScrollPane treeScroller=new JScrollPane(theTree);
	add(BorderLayout.CENTER,treeScroller);

	// add the rotation panel
	rotView=DataViewFactory.getDataRepresentationView(model.getDriverFactory().getRotationEncoding());
	
	// Select the root node
	theTree.setSelectionRow(0);
    }

    public Component getControlComponent()
    {
	return rotView.getComponent();
    }
    
    public Component getComponent()
    {
	return this;
    }

    public void initView() 
    { 
	// Initialize the panels
	rotView.initView();
    }

    /** Just call repaint */
    public void refresh()
    {
	if (currentHandle==null)
	    return;
	// Update the table
	rotView.refresh(currentHandle.getDataManager().getRotationRep(currentJoint));
    }
    
    private void initTreeNodes()
    {
	// Descend the model
	DefaultMutableTreeNode root=new DefaultMutableTreeNode();
	root.setUserObject(new Pair(null,null));
	
	for(Iterator i=model.getHandles().iterator();i.hasNext();) {
	    SkeletonHandle handle=(SkeletonHandle)i.next();
	    root.add(buildTreeModel(handle,handle.getRootJoint()));
	}
	
	// Create the tree model
	theTree=new JTree(new DefaultTreeModel(root));
	theTree.setVisibleRowCount(10);
	theTree.putClientProperty("JTree.lineStyle", "Angled");
	
	// Set up the selection model
	DefaultTreeSelectionModel treeSelectionModel=new DefaultTreeSelectionModel();
	treeSelectionModel.setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
	theTree.setSelectionModel(treeSelectionModel);
	theTree.addTreeSelectionListener(this);

	// Set the renderer
	JointNodeRenderer renderer=new JointNodeRenderer();
	renderer.setClosedIcon(renderer.getDefaultLeafIcon());
	renderer.setOpenIcon(renderer.getDefaultLeafIcon());
	theTree.setCellRenderer(renderer);

	// Make sure all rows are expanded
	for(int i=0;i<theTree.getRowCount();i++) {
	    theTree.expandRow(i);
	}
	
    }

    /**
     * Descend the joint hierarchy  and transform it to
     * a tree node hierarhcy.
     *
     * @param handle The skeleton that the joint belongs to
     * @param joint The joint.
     * @return The tree node.
     */
    private DefaultMutableTreeNode buildTreeModel(SkeletonHandle handle,Joint joint)
    {
	DefaultMutableTreeNode result=new DefaultMutableTreeNode();
	result.setUserObject(new Pair(joint,handle));
	
	// Descend
	for(Iterator i=joint.getChildren().iterator();i.hasNext();) {
	    //
	    Joint childJoint=(Joint)i.next();
	    // Recurse
	    DefaultMutableTreeNode childNode=buildTreeModel(handle,childJoint);
	    // Add it to the tree
	    result.add(childNode);
	}

	return result;
    }

    /**
     * The user has selected a joint. Update the channel table
     * accordingly.
     *
     * @param e The selection event.
     */
    public void valueChanged(TreeSelectionEvent e) 
    {
	// Get the path of the node
	TreePath nodePath=e.getPath();
	// Get the node
	DefaultMutableTreeNode treeNode=(DefaultMutableTreeNode)nodePath.getLastPathComponent();
	// Get teh pair
	Pair p=(Pair)treeNode.getUserObject();
	// Get the joint
	currentJoint=(Joint)p.first;
	// Get the data manager
	currentHandle=(SkeletonHandle)p.second;
	// Update the table
	refresh();
    }
}

/**
 * A renderer for joints
 */
class JointNodeRenderer extends DefaultTreeCellRenderer
{
    public Component getTreeCellRendererComponent(JTree tree, Object value, boolean sel, boolean expanded, boolean leaf, int row, boolean hasFocus) 
    {
	super.getTreeCellRendererComponent(tree,value,sel,expanded,leaf,row,hasFocus);
	// Get the node
	DefaultMutableTreeNode node=(DefaultMutableTreeNode)value;
	// Get the pair
	Pair p=(Pair)node.getUserObject();
	// Get the user object
	Joint joint=(Joint)p.first;
	// Set the text
	if (joint==null) {
	    setText("Skeleton");
	} else {
	    setText(joint.getName());
	}
	return this;
    }
}

